from __future__ import annotations

from typing import Optional

from nsj_rest_lib2.compiler.compiler_structures import PropertiesCompilerStructure
from nsj_rest_lib2.compiler.edl_model.api_model import HandlerConfig, HandlerMapping
from nsj_rest_lib2.compiler.edl_model.entity_model_base import EntityModelBase
from nsj_rest_lib2.compiler.edl_model.primitives import PrimitiveTypes
from nsj_rest_lib2.compiler.function_model import FunctionCompilationOutput
from nsj_rest_lib2.compiler.util.str_util import CompilerStrUtil
from nsj_rest_lib2.compiler.util.type_naming_util import compile_function_class_name
from nsj_rest_lib2.compiler.util.type_util import TypeUtil


class FunctionGetDeleteCompiler:
    """
    Responsável por gerar FunctionTypes (Get/List/Delete) baseados no
    binding de pg_function declarado no EDL.
    """

    def compile(
        self,
        entity_model: EntityModelBase,
        properties_structure: PropertiesCompilerStructure,
        handler_config: Optional[HandlerConfig],
        prefx_class_name: str,
        verb: str,
    ) -> FunctionCompilationOutput:
        if not isinstance(handler_config, HandlerConfig):
            return FunctionCompilationOutput()
        if handler_config.impl != "pg_function":
            return FunctionCompilationOutput()
        if not handler_config.call or not handler_config.call.arg_binding:
            return FunctionCompilationOutput()

        arg_binding = handler_config.call.arg_binding
        if not arg_binding.mapping or not arg_binding.type_name:
            return FunctionCompilationOutput()

        function_name = handler_config.function_ref
        if not function_name:
            return FunctionCompilationOutput()

        class_name = compile_function_class_name(
            entity_model.id,
            prefx_class_name,
            [],
            verb,
        )

        fields = self._build_fields(arg_binding.mapping, properties_structure)
        if not fields:
            return FunctionCompilationOutput()

        code = self._build_code(class_name, arg_binding.type_name, fields, verb)

        return FunctionCompilationOutput(
            class_name=class_name,
            code=code,
            function_name=function_name,
        )

    # ------------------------------------------------------------------ #
    # Helpers
    # ------------------------------------------------------------------ #

    def _build_fields(
        self,
        mappings: list[HandlerMapping],
        properties_structure: PropertiesCompilerStructure,
    ) -> list[tuple[str, str, Optional[str], bool]]:
        """
        Retorna lista de tuplas:
        (python_name, annotation, type_field_name, is_pk)
        """
        fields: list[tuple[str, str, Optional[str], bool]] = []
        used: set[str] = set()

        for entry in mappings or []:
            attr = (entry.attr or "").strip()
            if not attr:
                continue

            python_name = self._sanitize_identifier(attr)
            if python_name in used:
                continue
            used.add(python_name)

            type_field_name = attr if python_name != attr else None
            source = (entry.from_ or "").strip()
            is_pk = source.startswith("path.")

            annotation = self._resolve_annotation(entry, properties_structure)

            fields.append(
                (
                    python_name,
                    annotation,
                    type_field_name,
                    is_pk,
                )
            )

        return fields

    def _resolve_annotation(
        self,
        entry: HandlerMapping,
        properties_structure: PropertiesCompilerStructure,
    ) -> str:
        """
        Tenta inferir o tipo a partir das propriedades da entidade.
        Prioriza:
        - campos referenciados como body.<campo>;
        - ou, se o último segmento de 'from' (após o '.') existir como
          property do EDL, usa esse nome para lookup.
        """
        dto_field = self._extract_dto_field(entry.from_)

        # Se não veio de body.*, tenta casar o último segmento de 'from'
        # com o nome de alguma propriedade do EDL.
        if not dto_field:
            candidate = CompilerStrUtil.to_snake_case(
                (entry.from_ or "").split(".")[-1]
            )
            if candidate and properties_structure and candidate in properties_structure.properties:
                dto_field = candidate

        if not dto_field:
            return "Any"

        prop = (
            properties_structure.properties.get(dto_field)
            if properties_structure
            else None
        )
        if not prop or not isinstance(prop.type, PrimitiveTypes):
            return "Any"

        return TypeUtil.property_type_to_python_type(prop.type)

    def _build_code(
        self,
        class_name: str,
        type_name: str,
        fields: list[tuple[str, str, Optional[str], bool]],
        verb: str,
    ) -> str:
        needs_any = any(annotation == "Any" for _, annotation, _, _ in fields)
        needs_uuid = any(annotation == "uuid.UUID" for _, annotation, _, _ in fields)
        needs_datetime = any(
            annotation == "datetime.datetime" for _, annotation, _, _ in fields
        )

        lines: list[str] = []
        if needs_uuid:
            lines.append("import uuid")
        if needs_datetime:
            lines.append("import datetime")
        if needs_any:
            lines.append("from typing import Any")

        decorator_module, decorator_name, base_class = self._resolve_decorator(verb)
        lines.extend(
            [
                f"from {decorator_module} import {decorator_name}",
                f"from nsj_rest_lib.entity.function_type_base import {base_class}",
                "from nsj_rest_lib.descriptor.function_field import FunctionField",
                "",
                f"@{decorator_name}(type_name=\"{type_name}\")",
                f"class {class_name}({base_class}):",
            ]
        )

        for name, annotation, type_field_name, is_pk in fields:
            kwargs: list[str] = []
            if type_field_name:
                kwargs.append(f'type_field_name="{type_field_name}"')
            if is_pk:
                kwargs.append("pk=True")
            kwargs_str = ", ".join(kwargs)
            if kwargs_str:
                kwargs_str = ", " + kwargs_str
            lines.append(
                f"    {name}: {annotation} = FunctionField({kwargs_str.lstrip(', ')})"
            )

        if not fields:
            lines.append("    pass")

        lines.append("")  # newline final
        return "\n".join(lines)

    def _resolve_decorator(self, verb: str) -> tuple[str, str, str]:
        verb = (verb or "").lower()
        if verb == "get":
            return (
                "nsj_rest_lib.decorator.get_function_type",
                "GetFunctionType",
                "GetFunctionTypeBase",
            )
        if verb == "list":
            return (
                "nsj_rest_lib.decorator.list_function_type",
                "ListFunctionType",
                "ListFunctionTypeBase",
            )
        if verb == "delete":
            return (
                "nsj_rest_lib.decorator.delete_function_type",
                "DeleteFunctionType",
                "DeleteFunctionTypeBase",
            )
        # fallback genérico
        return (
            "nsj_rest_lib.decorator.function_type",
            "FunctionType",
            "FunctionTypeBase",
        )

    def _sanitize_identifier(self, name: str) -> str:
        candidate = CompilerStrUtil.to_snake_case(name)
        candidate = "".join(ch if ch.isalnum() or ch == "_" else "_" for ch in candidate)
        if not candidate:
            candidate = "field"
        if candidate[0].isdigit():
            candidate = f"_{candidate}"
        return candidate

    def _extract_dto_field(self, source: Optional[str]) -> Optional[str]:
        if not source:
            return None

        if source.startswith("body."):
            segment = source[5:].split(".", 1)[0]
            return CompilerStrUtil.to_snake_case(segment)

        return None



import re
from dataclasses import dataclass
from typing import Literal, Sequence

from nsj_rest_lib2.compiler.edl_model.entity_model_base import EntityModelBase
from nsj_rest_lib2.compiler.edl_model.primitives import (
    REGEX_EXTERNAL_COMPONENT_REF,
    REGEX_EXTERNAL_REF,
    REGEX_INTERNAL_REF,
)


RelationRefType = Literal["internal", "external", "external_component"]


@dataclass
class RelationRef:
    ref_type: RelationRefType
    scope: str | None
    entity: str
    components: list[str]

    @property
    def entity_key(self) -> str | None:
        if not self.scope:
            return None

        return f"{self.scope}/{self.entity}"

    @property
    def target_id(self) -> str:
        return self.components[-1] if self.components else self.entity

    @property
    def prefx_class_name(self) -> str:
        if not self.components:
            return ""

        prefix_parts: list[str] = [self.entity, *self.components[:-1]]
        return f"_{'_'.join(prefix_parts)}"

    @property
    def is_external(self) -> bool:
        return self.ref_type in {"external", "external_component"}

    def resolve_model(
        self,
        entity_models: dict[str, EntityModelBase],
        base_model: EntityModelBase | None = None,
    ) -> EntityModelBase | None:
        entity_key = self.entity_key
        resolved_base: EntityModelBase | None = None

        if entity_key:
            resolved_base = entity_models.get(entity_key)
        elif base_model:
            resolved_base = base_model

        if not resolved_base:
            return None

        return RelationRefParser.follow_components(resolved_base, self.components)


class RelationRefParser:
    @staticmethod
    def parse(ref: str) -> RelationRef | None:
        internal_match = re.match(REGEX_INTERNAL_REF, ref)
        if internal_match:
            components = RelationRefParser._split_components(internal_match.group(1))
            if not components:
                return None
            entity = components[0]
            extra_components = components[1:] if len(components) > 1 else []
            return RelationRef("internal", None, entity, extra_components)

        external_component_match = re.match(REGEX_EXTERNAL_COMPONENT_REF, ref)
        if external_component_match:
            components = RelationRefParser._split_components(
                external_component_match.group(3)
            )
            return RelationRef(
                "external_component",
                external_component_match.group(1),
                external_component_match.group(2),
                components,
            )

        external_match = re.match(REGEX_EXTERNAL_REF, ref)
        if external_match:
            return RelationRef(
                "external",
                external_match.group(1),
                external_match.group(2),
                [],
            )

        return None

    @staticmethod
    def follow_components(
        base_model: EntityModelBase, components: Sequence[str]
    ) -> EntityModelBase | None:
        target_model: EntityModelBase | None = base_model
        for component in components:
            if not target_model or not target_model.components:
                return None
            target_model = target_model.components.get(component)
            if target_model is None:
                return None

        return target_model

    @staticmethod
    def _split_components(components_path: str) -> list[str]:
        # Remove marcadores opcionais '#/components/' intermediários
        cleaned = components_path.replace("#/components/", "/").replace("#", "")
        return [part for part in cleaned.split("/") if part]



from __future__ import annotations

import ast
from typing import Any, Optional

from nsj_rest_lib2.compiler.dto_compiler import DTOCompiler
from nsj_rest_lib2.compiler.edl_model.entity_model_base import EntityModelBase
from nsj_rest_lib2.compiler.edl_model.primitives import BasicTypes
from nsj_rest_lib2.compiler.util.str_util import CompilerStrUtil
from nsj_rest_lib2.compiler.util.type_naming_util import (
    compile_response_dto_class_name,
)


class ResponseDTOCompiler:
    def __init__(
        self,
        dto_compiler: DTOCompiler,
        entity_model: EntityModelBase,
        prefx_class_name: str,
        ast_dto_attributes: list[ast.stmt],
        aux_classes: list[ast.stmt],
        related_imports: list[tuple[str, str, str]],
        fixed_filters: list[tuple[str, BasicTypes]],
    ) -> None:
        self._dto_compiler = dto_compiler
        self._entity_model = entity_model
        self._prefx_class_name = prefx_class_name
        self._aux_classes = aux_classes
        self._related_imports = related_imports
        self._fixed_filters = fixed_filters
        self._dto_attr_map = self._build_attr_map(ast_dto_attributes)

    @staticmethod
    def handler_result_details(handler: Any) -> tuple[str, list[str] | None]:
        expected = "empty"
        properties = None
        if handler and getattr(handler, "result", None):
            expected = handler.result.expected or "empty"
            properties = handler.result.properties
        return (expected, properties)

    def compile_partial_response_dto(
        self, verb: str, properties: list[str]
    ) -> tuple[str, str]:
        if not properties:
            raise Exception(
                f"result.properties não pode ser vazio para '{verb}' na entidade '{self._entity_model.id}'."
            )

        normalized = [
            CompilerStrUtil.to_snake_case(prop) for prop in properties
        ]
        missing = [
            properties[idx]
            for idx, norm in enumerate(normalized)
            if norm not in self._dto_attr_map
        ]
        if missing:
            raise Exception(
                f"As propriedades {missing} não foram encontradas na entidade '{self._entity_model.id}'."
            )

        filtered_attrs = [self._dto_attr_map[name] for name in normalized]
        filtered_fixed_filters = [
            item for item in self._fixed_filters if item[0] in normalized
        ]

        class_name = compile_response_dto_class_name(
            self._entity_model.id,
            self._prefx_class_name,
            verb,
        )

        return self._dto_compiler.compile(
            self._entity_model,
            filtered_attrs,
            self._aux_classes,
            self._related_imports,
            filtered_fixed_filters,
            self._prefx_class_name,
            None,
            class_name_override=class_name,
        )

    @staticmethod
    def _extract_attr_name(stmt: ast.stmt) -> Optional[str]:
        if isinstance(stmt, ast.AnnAssign) and isinstance(stmt.target, ast.Name):
            return stmt.target.id
        if isinstance(stmt, ast.Assign) and stmt.targets:
            target = stmt.targets[0]
            if isinstance(target, ast.Name):
                return target.id
        return None

    def _build_attr_map(
        self, ast_dto_attributes: list[ast.stmt]
    ) -> dict[str, ast.stmt]:
        dto_attr_map: dict[str, ast.stmt] = {}
        for stmt in ast_dto_attributes:
            name = self._extract_attr_name(stmt)
            if name:
                dto_attr_map[name] = stmt
        return dto_attr_map



import datetime
import json
import os
import re
import sys
import threading
import types
from typing import Any

from nsj_rest_lib.settings import get_logger

import yaml

from nsj_rest_lib2.compiler.compiler import EDLCompiler
from nsj_rest_lib2.compiler.edl_model.entity_model import EntityModel
from nsj_rest_lib2.compiler.edl_model.entity_model_base import EntityModelBase
from nsj_rest_lib2.compiler.edl_model.entity_model_root import EntityModelRoot
from nsj_rest_lib2.compiler.edl_model.primitives import REGEX_EXTERNAL_REF
from nsj_rest_lib2.compiler.model import CompilerResult, RelationDependency
from nsj_rest_lib2.compiler.util.type_naming_util import (
    build_entity_hash as build_entity_hash_util,
    compile_namespace_keys,
    resolve_namespace_key as resolve_namespace_key_util,
)
from nsj_rest_lib2.dto.escopo_dto import EscopoDTO
from nsj_rest_lib2.exception import MissingEntityConfigException
from nsj_rest_lib2.settings import (
    EDLS_FROM_REDIS,
    ESCOPO_RESTLIB2,
    MIN_TIME_SOURCE_REFRESH,
)


class LoadedEntity:
    def __init__(self):
        self.dto_class_name: str = ""
        self.entity_class_name: str = ""
        self.entity_hash: str = ""
        self.loaded_at: datetime.datetime = datetime.datetime.now()
        self.api_expose: bool = False
        self.api_verbs: list[str] = []
        self.relations_dependencies: list[RelationDependency] = []
        self.never_expire: bool = False
        self.service_account: str | None = None
        self.insert_function_class_name: str | None = None
        self.update_function_class_name: str | None = None
        self.insert_function_name: str | None = None
        self.update_function_name: str | None = None
        self.get_function_name: str | None = None
        self.list_function_name: str | None = None
        self.delete_function_name: str | None = None
        self.get_function_type_class_name: str | None = None
        self.list_function_type_class_name: str | None = None
        self.delete_function_type_class_name: str | None = None
        self.retrieve_after_insert: bool = False
        self.retrieve_after_update: bool = False
        self.retrieve_after_partial_update: bool = False
        self.post_response_dto_class_name: str | None = None
        self.put_response_dto_class_name: str | None = None
        self.patch_response_dto_class_name: str | None = None
        self.custom_json_post_response: bool = False
        self.custom_json_put_response: bool = False
        self.custom_json_patch_response: bool = False
        self.custom_json_get_response: bool = False
        self.custom_json_list_response: bool = False
        self.custom_json_delete_response: bool = False


class Namespace:
    def __init__(self):
        self.key: str = ""
        self.loaded_entities: dict[str, LoadedEntity] = {}
        self.entities_dict: dict = {}
        self.module: types.ModuleType = types.ModuleType("empty")


namespaces_dict: dict[str, Namespace] = {}


class EntityLoader:
    def __init__(self, edls_path: str | None = None) -> types.NoneType:
        self._lock = threading.Lock()
        if edls_path:
            self._load_edls_from_disk(edls_path)

    def load_entity_source(
        self,
        entity_resource: str,
        tenant: str | None,
        grupo_empresarial: str | None,
        escopo: str = ESCOPO_RESTLIB2,
        force_reload: bool = False,
    ) -> tuple[
        str,
        str,
        dict,
        bool,
        list[str],
        str | None,
        str | None,
        str | None,
        str | None,
        str | None,
        str | None,
        str | None,
        str | None,
        str | None,
        str | None,
        str | None,
        bool,
        bool,
        bool,
        str | None,
        str | None,
        str | None,
        bool,
        bool,
        bool,
        bool,
        bool,
        bool,
    ]:
        # Assumind o escopo default se necessário
        if not escopo:
            escopo = ESCOPO_RESTLIB2

        # Montando as chaves dos namespaces
        grupo_key, tenant_key, default_key = compile_namespace_keys(
            tenant, grupo_empresarial
        )

        result = self._search_entity_namespace_in_memory(
            entity_resource, grupo_key, tenant_key, default_key
        )

        # Se conseguiu localizar na memória, verifica se houve alteração no hash, em relação ao redis
        if result is not None:
            # Desempacotando o result e recuperando informações do namespace
            (
                entity_config_key,
                namespace,
            ) = result

            loaded_entity = namespace.loaded_entities[entity_resource]
            dto_class_name = loaded_entity.dto_class_name
            entity_class_name = loaded_entity.entity_class_name
            entities_dict = namespace.entities_dict
            api_expose = loaded_entity.api_expose
            api_verbs = loaded_entity.api_verbs
            service_account = loaded_entity.service_account
            relations_dependencies = loaded_entity.relations_dependencies
            insert_function_class_name = loaded_entity.insert_function_class_name
            update_function_class_name = loaded_entity.update_function_class_name
            insert_function_name = loaded_entity.insert_function_name
            update_function_name = loaded_entity.update_function_name
            get_function_name = loaded_entity.get_function_name
            list_function_name = loaded_entity.list_function_name
            delete_function_name = loaded_entity.delete_function_name
            get_function_type_class_name = (
                loaded_entity.get_function_type_class_name
            )
            list_function_type_class_name = (
                loaded_entity.list_function_type_class_name
            )
            delete_function_type_class_name = (
                loaded_entity.delete_function_type_class_name
            )

            cached_result = (
                dto_class_name,
                entity_class_name,
                entities_dict,
                api_expose,
                api_verbs,
                service_account,
                insert_function_class_name,
                update_function_class_name,
                insert_function_name,
                update_function_name,
                get_function_name,
                list_function_name,
                delete_function_name,
                get_function_type_class_name,
                list_function_type_class_name,
                delete_function_type_class_name,
                loaded_entity.retrieve_after_insert,
                loaded_entity.retrieve_after_update,
                loaded_entity.retrieve_after_partial_update,
                loaded_entity.post_response_dto_class_name,
                loaded_entity.put_response_dto_class_name,
                loaded_entity.patch_response_dto_class_name,
                loaded_entity.custom_json_post_response,
                loaded_entity.custom_json_put_response,
                loaded_entity.custom_json_patch_response,
                loaded_entity.custom_json_get_response,
                loaded_entity.custom_json_list_response,
                loaded_entity.custom_json_delete_response,
            )

            # Verificando se alguma de suas dependências precisariam ser recarregadas
            for rd in relations_dependencies:
                if rd.entity_resource is None or rd.entity_scope is None:
                    raise RuntimeError(
                        f"Erro: Dependência de entidade mal formada na entidade {entity_resource}."
                    )

                self.load_entity_source(
                    rd.entity_resource,
                    str(rd.tenant),
                    str(rd.grupo_empresarial),
                    rd.entity_scope,
                    force_reload=force_reload,
                )

            if loaded_entity.never_expire or not EDLS_FROM_REDIS:
                return cached_result

            # Se o tempo entre o carregamento e agora for maior do que MIN_TIME_SOURCE_REFRESH minutos,
            # verifica se precisa de refresh
            time_diff = datetime.datetime.now() - loaded_entity.loaded_at

            if (
                time_diff.total_seconds() >= MIN_TIME_SOURCE_REFRESH * 60
                or force_reload
            ):
                # Renovando o tempo de refresh
                loaded_entity.loaded_at = datetime.datetime.now()

                # Recuperando do Redis direto pela key (faz uma só chamada ao redis)
                loaded_config = self._load_entity_config_from_redis(
                    entity_resource,
                    grupo_key,
                    tenant_key,
                    default_key,
                    entity_config_key,
                    escopo=escopo,
                )

                # Se não achar no redis, usa o que estava em memória
                if not loaded_config:
                    return cached_result

                # Desempacotando resultado
                entity_config_key, entity_config_str = loaded_config

                # Executando o código da entidade, só se houver mudança no hash
                result_execute = self._execute_entity_source(
                    entity_config_str,
                    entity_config_key,
                    entity_resource,
                    check_refresh=True,
                )

                # Se não carregou novo código, usa o que estava em memória
                if result_execute is None:
                    return cached_result
                else:
                    (
                        dto_class_name,
                        entity_class_name,
                        namespace,
                        api_expose,
                        api_verbs,
                        service_account,
                        insert_function_class_name,
                        update_function_class_name,
                        insert_function_name,
                        update_function_name,
                        get_function_name,
                        list_function_name,
                        delete_function_name,
                        get_function_type_class_name,
                        list_function_type_class_name,
                        delete_function_type_class_name,
                        retrieve_after_insert,
                        retrieve_after_update,
                        retrieve_after_partial_update,
                        post_response_dto_class_name,
                        put_response_dto_class_name,
                        patch_response_dto_class_name,
                        custom_json_post_response,
                        custom_json_put_response,
                        custom_json_patch_response,
                        custom_json_get_response,
                        custom_json_list_response,
                        custom_json_delete_response,
                    ) = result_execute
                    return (
                        dto_class_name,
                        entity_class_name,
                        namespace.entities_dict,
                        api_expose,
                        api_verbs,
                        service_account,
                        insert_function_class_name,
                        update_function_class_name,
                        insert_function_name,
                        update_function_name,
                        get_function_name,
                        list_function_name,
                        delete_function_name,
                        get_function_type_class_name,
                        list_function_type_class_name,
                        delete_function_type_class_name,
                        retrieve_after_insert,
                        retrieve_after_update,
                        retrieve_after_partial_update,
                        post_response_dto_class_name,
                        put_response_dto_class_name,
                        patch_response_dto_class_name,
                        custom_json_post_response,
                        custom_json_put_response,
                        custom_json_patch_response,
                        custom_json_get_response,
                        custom_json_list_response,
                        custom_json_delete_response,
                    )
            else:
                # Se não deu o intervalo de verificação do refresh, retorna o que está em memória
                return cached_result

        if not EDLS_FROM_REDIS:
            raise MissingEntityConfigException()

        # Se não conseguir recuperar a entidade, procura no redis:
        loaded_config = self._load_entity_config_from_redis(
            entity_resource,
            grupo_key,
            tenant_key,
            default_key,
            None,
            escopo=escopo,
        )

        # Se também não achar no redis, lanca exceção
        if not loaded_config:
            raise MissingEntityConfigException()

        # Desempacotando resultado
        entity_config_key, entity_config_str = loaded_config

        # Executando o código da entidade
        result_execute = self._execute_entity_source(
            entity_config_str, entity_config_key, entity_resource
        )

        if result_execute is None:
            raise RuntimeError(
                f"Erro desconhecido carregando entidade: {entity_resource}"
            )
        (
            dto_class_name,
            entity_class_name,
            namespace,
            api_expose,
            api_verbs,
            service_account,
            insert_function_class_name,
            update_function_class_name,
            insert_function_name,
            update_function_name,
            get_function_name,
            list_function_name,
            delete_function_name,
            get_function_type_class_name,
            list_function_type_class_name,
            delete_function_type_class_name,
            retrieve_after_insert,
            retrieve_after_update,
            retrieve_after_partial_update,
            post_response_dto_class_name,
            put_response_dto_class_name,
            patch_response_dto_class_name,
            custom_json_post_response,
            custom_json_put_response,
            custom_json_patch_response,
            custom_json_get_response,
            custom_json_list_response,
            custom_json_delete_response,
        ) = result_execute

        return (
            dto_class_name,
            entity_class_name,
            namespace.entities_dict,
            api_expose,
            api_verbs,
            service_account,
            insert_function_class_name,
            update_function_class_name,
            insert_function_name,
            update_function_name,
            get_function_name,
            list_function_name,
            delete_function_name,
            get_function_type_class_name,
            list_function_type_class_name,
            delete_function_type_class_name,
            retrieve_after_insert,
            retrieve_after_update,
            retrieve_after_partial_update,
            post_response_dto_class_name,
            put_response_dto_class_name,
            patch_response_dto_class_name,
            custom_json_post_response,
            custom_json_put_response,
            custom_json_patch_response,
            custom_json_get_response,
            custom_json_list_response,
            custom_json_delete_response,
        )

    def clear_namespaces(self):
        """
        Clears all loaded namespaces from memory.

        This method removes all entries from the namespaces_dict, effectively resetting
        the in-memory cache of loaded entities and their associated namespaces.
        """
        with self._lock:
            namespaces_dict.clear()

    def _load_edls_from_disk(self, edls_path: str) -> None:
        if namespaces_dict:
            return

        resolved_path = self._resolve_edls_path(edls_path)
        if not resolved_path:
            return

        if os.path.isdir(resolved_path):
            edl_files = self._list_edl_files(resolved_path)
        elif os.path.isfile(resolved_path) and self._is_edl_file(resolved_path):
            edl_files = [resolved_path]
        else:
            return

        if not edl_files:
            return

        entities: dict[str, EntityModelBase] = {}
        for file_path in edl_files:
            edl_json = self._read_edl_file(file_path)
            if not isinstance(edl_json, dict):
                continue

            if edl_json.get("mixin", False):
                entity_model = EntityModelRoot(**edl_json)
            else:
                entity_model = EntityModel(**edl_json)

            complete_entity_id = f"{entity_model.escopo}/{entity_model.id}"
            entities[complete_entity_id] = entity_model

        if not entities:
            return

        escopos: dict[str, EscopoDTO] = {}
        for entity_model in entities.values():
            if entity_model.escopo not in escopos:
                escopos[entity_model.escopo] = EscopoDTO(
                    codigo=entity_model.escopo,
                    service_account=None,
                )

        compiler = EDLCompiler()
        dependencies = list(entities.items())

        for entity_model in entities.values():
            escopo_dto = escopos.get(entity_model.escopo)
            if escopo_dto is None:
                continue

            compiler_result = compiler.compile_model(
                entity_model,
                dependencies,
                escopo=escopo_dto,
            )
            if not compiler_result:
                continue

            entity_config = self._build_entity_payload(
                entity_model,
                compiler_result,
            )
            entity_resource = entity_config.get("api_resource")
            if not entity_resource:
                continue

            entity_config_key = self._resolve_namespace_key(entity_model)
            self._execute_entity_source(
                json.dumps(entity_config, ensure_ascii=False),
                entity_config_key,
                entity_resource,
                from_disk=True,
                skip_dependency_load=not EDLS_FROM_REDIS,
            )

    def _resolve_edls_path(self, edls_path: str) -> str | None:
        edls_path = edls_path.strip()
        if not edls_path:
            return None
        if os.path.isabs(edls_path):
            return edls_path
        return os.path.join(os.getcwd(), edls_path)

    def _is_edl_file(self, file_path: str) -> bool:
        lower_name = file_path.lower()
        return lower_name.endswith(".json") or lower_name.endswith(
            ".yml"
        ) or lower_name.endswith(".yaml")

    def _list_edl_files(self, root_path: str) -> list[str]:
        edl_files: list[str] = []
        for current_root, _, files in os.walk(root_path):
            for file_name in files:
                file_path = os.path.join(current_root, file_name)
                if self._is_edl_file(file_path):
                    edl_files.append(file_path)
        edl_files.sort()
        return edl_files

    def _read_edl_file(self, file_path: str) -> dict[str, Any] | None:
        with open(file_path, "r", encoding="utf-8") as file_handler:
            if file_path.lower().endswith(".json"):
                return json.load(file_handler)
            return yaml.safe_load(file_handler)

    def _resolve_namespace_key(self, entity_model: EntityModelBase) -> str:
        tenant = getattr(entity_model, "tenant", None)
        grupo_empresarial = getattr(entity_model, "grupo_empresarial", None)

        return resolve_namespace_key_util(tenant, grupo_empresarial)

    def _build_entity_payload(
        self,
        entity_model: EntityModelBase,
        compiler_result: CompilerResult,
    ) -> dict[str, Any]:
        api_resource = compiler_result.api_resource
        if not api_resource:
            if isinstance(entity_model, EntityModel) and entity_model.api:
                api_resource = entity_model.api.resource
        if not api_resource:
            raise ValueError("api.resource é obrigatório para carregar EDLs.")

        relations = [
            rd.to_dict()
            for rd in (compiler_result.relations_dependencies or [])
        ]

        return {
            "dto_class_name": compiler_result.dto_class_name,
            "entity_class_name": compiler_result.entity_class_name,
            "service_account": compiler_result.service_account,
            "insert_function_class_name": compiler_result.insert_function_class_name,
            "insert_function_name": compiler_result.insert_function_name,
            "source_insert_function": compiler_result.source_insert_function,
            "update_function_class_name": compiler_result.update_function_class_name,
            "update_function_name": compiler_result.update_function_name,
            "source_update_function": compiler_result.source_update_function,
            "get_function_name": compiler_result.get_function_name,
            "list_function_name": compiler_result.list_function_name,
            "delete_function_name": compiler_result.delete_function_name,
            "get_function_type_class_name": compiler_result.get_function_type_class_name,
            "list_function_type_class_name": compiler_result.list_function_type_class_name,
            "delete_function_type_class_name": compiler_result.delete_function_type_class_name,
            "source_get_function_type": compiler_result.source_get_function_type,
            "source_list_function_type": compiler_result.source_list_function_type,
            "source_delete_function_type": compiler_result.source_delete_function_type,
            "retrieve_after_insert": compiler_result.retrieve_after_insert,
            "retrieve_after_update": compiler_result.retrieve_after_update,
            "retrieve_after_partial_update": compiler_result.retrieve_after_partial_update,
            "post_response_dto_class_name": compiler_result.post_response_dto_class_name,
            "put_response_dto_class_name": compiler_result.put_response_dto_class_name,
            "patch_response_dto_class_name": compiler_result.patch_response_dto_class_name,
            "custom_json_post_response": compiler_result.custom_json_post_response,
            "custom_json_put_response": compiler_result.custom_json_put_response,
            "custom_json_patch_response": compiler_result.custom_json_patch_response,
            "custom_json_get_response": compiler_result.custom_json_get_response,
            "custom_json_list_response": compiler_result.custom_json_list_response,
            "custom_json_delete_response": compiler_result.custom_json_delete_response,
            "source_dto": compiler_result.dto_code,
            "source_entity": compiler_result.entity_code,
            "entity_hash": self._build_entity_hash(compiler_result),
            "api_expose": compiler_result.api_expose,
            "api_resource": api_resource,
            "api_verbs": compiler_result.api_verbs or [],
            "relations_dependencies": relations,
        }

    def _build_entity_hash(self, compiler_result: CompilerResult) -> str:
        return build_entity_hash_util(compiler_result)

    def _ensure_dynamic_package(self):
        """
        Garante que exista um pacote 'dynamic' em sys.modules.
        """
        pkg = sys.modules.get("dynamic")
        if pkg is None:
            pkg = types.ModuleType("dynamic")
            pkg.__path__ = []  # marca como pacote
            pkg.__package__ = "dynamic"
            sys.modules["dynamic"] = pkg
        return pkg

    def _execute_entity_source(
        self,
        entity_config_str: str,
        entity_config_key: str,
        entity_resource: str,
        check_refresh: bool = False,
        from_disk: bool = False,
        skip_dependency_load: bool = False,
    ) -> tuple[
        str,
        str,
        Namespace,
        bool,
        list[str],
        str | None,
        str | None,
        str | None,
        str | None,
        str | None,
        str | None,
        str | None,
        str | None,
        str | None,
        str | None,
        str | None,
        bool,
        bool,
        bool,
        str | None,
        str | None,
        str | None,
        bool,
        bool,
        bool,
        bool,
        bool,
        bool,
    ] | None:
        # Interpretando o json de configuração da entidade
        try:
            entity_config = json.loads(entity_config_str)

            dto_class_name = entity_config["dto_class_name"]
            entity_class_name = entity_config["entity_class_name"]
            source_dto = entity_config["source_dto"]
            source_entity = entity_config["source_entity"]
            entity_hash = entity_config["entity_hash"]

            api_expose = entity_config["api_expose"]
            # api_resource = entity_config["api_resource"]
            api_verbs = entity_config["api_verbs"]
            service_account = entity_config.get("service_account")
            insert_function_class_name = entity_config.get(
                "insert_function_class_name"
            )
            insert_function_name = entity_config.get("insert_function_name")
            insert_function_code = entity_config.get("source_insert_function")
            update_function_class_name = entity_config.get(
                "update_function_class_name"
            )
            update_function_name = entity_config.get("update_function_name")
            update_function_code = entity_config.get("source_update_function")
            get_function_name = entity_config.get("get_function_name")
            list_function_name = entity_config.get("list_function_name")
            delete_function_name = entity_config.get("delete_function_name")
            get_function_type_class_name = entity_config.get(
                "get_function_type_class_name"
            )
            list_function_type_class_name = entity_config.get(
                "list_function_type_class_name"
            )
            delete_function_type_class_name = entity_config.get(
                "delete_function_type_class_name"
            )
            retrieve_after_insert = entity_config.get(
                "retrieve_after_insert", False
            )
            retrieve_after_update = entity_config.get(
                "retrieve_after_update", False
            )
            retrieve_after_partial_update = entity_config.get(
                "retrieve_after_partial_update", False
            )
            post_response_dto_class_name = entity_config.get(
                "post_response_dto_class_name"
            )
            put_response_dto_class_name = entity_config.get(
                "put_response_dto_class_name"
            )
            patch_response_dto_class_name = entity_config.get(
                "patch_response_dto_class_name"
            )
            custom_json_post_response = entity_config.get(
                "custom_json_post_response", False
            )
            custom_json_put_response = entity_config.get(
                "custom_json_put_response", False
            )
            custom_json_patch_response = entity_config.get(
                "custom_json_patch_response", False
            )
            custom_json_get_response = entity_config.get(
                "custom_json_get_response", False
            )
            custom_json_list_response = entity_config.get(
                "custom_json_list_response", False
            )
            custom_json_delete_response = entity_config.get(
                "custom_json_delete_response", False
            )
            get_function_code = entity_config.get("source_get_function_type")
            list_function_code = entity_config.get("source_list_function_type")
            delete_function_code = entity_config.get("source_delete_function_type")
            relations_dependencies = [
                RelationDependency().from_dict(rd)
                for rd in entity_config.get("relations_dependencies", [])
            ]
        except json.JSONDecodeError as e:
            if not check_refresh:
                raise RuntimeError(
                    f"Erro ao decodificar JSON da entidade {entity_resource}; na chave {entity_config_key}: {e}"
                )
            else:
                get_logger().error(
                    f"Erro ao decodificar JSON da entidade {entity_resource}; na chave {entity_config_key}: {e}"
                )
                return None

        # Verificando se alguma de suas dependências precisariam ser carregadas (ou recarregadas)
        if not skip_dependency_load:
            for rd in relations_dependencies:
                if rd.entity_resource is None or rd.entity_scope is None:
                    raise RuntimeError(
                        f"Erro: Dependência de entidade mal formada na entidade {entity_resource}."
                    )

                self.load_entity_source(
                    rd.entity_resource,
                    str(rd.tenant),
                    str(rd.grupo_empresarial),
                    rd.entity_scope,
                )

        # Verificando se a entidade precisa ou não de refresh
        if check_refresh:
            loaded_namespace = namespaces_dict.get(entity_config_key)
            if not loaded_namespace:
                return None

            loaded_entity = loaded_namespace.loaded_entities.get(entity_resource)
            if not loaded_entity:
                return None

            if loaded_entity.entity_hash == entity_hash:
                return None

        # Imprimindo alerta de load no log
        get_logger().debug(
            f"Carregando entidade {entity_resource} no namespace {entity_config_key}."
        )

        # Carregando a entidade no namespace
        with self._lock:
            self._ensure_dynamic_package()

            namespace = namespaces_dict.get(entity_config_key)
            if namespace is None:
                namespace = Namespace()
                namespace.key = entity_config_key
                namespaces_dict[entity_config_key] = namespace

            # Hot reload: removendo o módulo do sys.modules, se existir
            full_name = f"dynamic.{entity_config_key}"
            # if full_name in sys.modules:
            #     sys.modules.pop(full_name)

            # Executando o código da entidade
            module = sys.modules.get(full_name)
            if not module:
                module = types.ModuleType(full_name)
                module.__package__ = "dynamic"
                module.__dict__["__builtins__"] = __builtins__
                sys.modules[full_name] = module

                parent = sys.modules["dynamic"]
                setattr(parent, entity_config_key, module)

                namespace.module = module
                namespace.entities_dict = module.__dict__

            get_logger().debug(
                f"Executando o código da entidade {entity_resource} no namespace {entity_config_key}. Código:"
            )
            get_logger().debug(f"Entity source:\n{source_entity}")
            get_logger().debug(f"DTO source:\n{source_dto}")

            if insert_function_code:
                self._safe_exec(
                    insert_function_code,
                    namespace.entities_dict,
                    "Insert Function source",
                )

            if update_function_code:
                self._safe_exec(
                    update_function_code,
                    namespace.entities_dict,
                    "Update Function source",
                )

            if get_function_code:
                self._safe_exec(
                    get_function_code,
                    namespace.entities_dict,
                    "Get FunctionType source",
                )

            if list_function_code:
                self._safe_exec(
                    list_function_code,
                    namespace.entities_dict,
                    "List FunctionType source",
                )

            if delete_function_code:
                self._safe_exec(
                    delete_function_code,
                    namespace.entities_dict,
                    "Delete FunctionType source",
                )

            self._safe_exec(source_entity, namespace.entities_dict, "Entity source")
            self._safe_exec(source_dto, namespace.entities_dict, "DTO source")

            # Gravando a entidade no dict de entidades carregadas
            loaded_entity = LoadedEntity()
            loaded_entity.dto_class_name = dto_class_name
            loaded_entity.entity_class_name = entity_class_name
            loaded_entity.entity_hash = entity_hash
            loaded_entity.api_expose = api_expose
            loaded_entity.api_verbs = api_verbs
            loaded_entity.service_account = service_account
            loaded_entity.relations_dependencies = relations_dependencies
            loaded_entity.never_expire = from_disk
            loaded_entity.insert_function_class_name = insert_function_class_name
            loaded_entity.update_function_class_name = update_function_class_name
            loaded_entity.insert_function_name = insert_function_name
            loaded_entity.update_function_name = update_function_name
            loaded_entity.get_function_name = get_function_name
            loaded_entity.list_function_name = list_function_name
            loaded_entity.delete_function_name = delete_function_name
            loaded_entity.get_function_type_class_name = (
                get_function_type_class_name
            )
            loaded_entity.list_function_type_class_name = (
                list_function_type_class_name
            )
            loaded_entity.delete_function_type_class_name = (
                delete_function_type_class_name
            )
            loaded_entity.retrieve_after_insert = bool(retrieve_after_insert)
            loaded_entity.retrieve_after_update = bool(retrieve_after_update)
            loaded_entity.retrieve_after_partial_update = bool(
                retrieve_after_partial_update
            )
            loaded_entity.post_response_dto_class_name = (
                post_response_dto_class_name
            )
            loaded_entity.put_response_dto_class_name = (
                put_response_dto_class_name
            )
            loaded_entity.patch_response_dto_class_name = (
                patch_response_dto_class_name
            )
            loaded_entity.custom_json_post_response = bool(
                custom_json_post_response
            )
            loaded_entity.custom_json_put_response = bool(
                custom_json_put_response
            )
            loaded_entity.custom_json_patch_response = bool(
                custom_json_patch_response
            )
            loaded_entity.custom_json_get_response = bool(
                custom_json_get_response
            )
            loaded_entity.custom_json_list_response = bool(
                custom_json_list_response
            )
            loaded_entity.custom_json_delete_response = bool(
                custom_json_delete_response
            )

            namespace.loaded_entities[entity_resource] = loaded_entity

        return (
            dto_class_name,
            entity_class_name,
            namespace,
            api_expose,
            api_verbs,
            service_account,
            insert_function_class_name,
            update_function_class_name,
            insert_function_name,
            update_function_name,
            get_function_name,
            list_function_name,
            delete_function_name,
            get_function_type_class_name,
            list_function_type_class_name,
            delete_function_type_class_name,
            bool(retrieve_after_insert),
            bool(retrieve_after_update),
            bool(retrieve_after_partial_update),
            post_response_dto_class_name,
            put_response_dto_class_name,
            patch_response_dto_class_name,
            bool(custom_json_post_response),
            bool(custom_json_put_response),
            bool(custom_json_patch_response),
            bool(custom_json_get_response),
            bool(custom_json_list_response),
            bool(custom_json_delete_response),
        )

    def _safe_exec(self, source_code, context, description):
        try:
            exec(source_code, context)
        except Exception as e:
            get_logger().error(f"Error executing {description}: {e}")
            raise

    def _load_entity_config_from_redis(
        self,
        entity_resource: str,
        grupo_key: str,
        tenant_key: str,
        default_key: str,
        entity_config_key: str | None,
        escopo: str,
    ) -> tuple[str, str] | None:
        if not EDLS_FROM_REDIS:
            return None

        from nsj_rest_lib2.redis_config import get_redis

        get_logger().debug(
            f"Procurando a configuração da entidade {entity_resource} no redis. Tenant key: {tenant_key} e Grupo key: {grupo_key}"
        )

        if entity_config_key is not None:
            entity_config_str = get_redis(
                "entity_config", escopo, entity_config_key, entity_resource
            )

        else:
            entity_config_key = grupo_key
            entity_config_str = get_redis(
                "entity_config", escopo, grupo_key, entity_resource
            )
            if entity_config_str is None:
                entity_config_key = tenant_key
                entity_config_str = get_redis(
                    "entity_config", escopo, tenant_key, entity_resource
                )
            if entity_config_str is None:
                entity_config_key = default_key
                entity_config_str = get_redis(
                    "entity_config", escopo, default_key, entity_resource
                )

        # Se não encontrar no redis, retorna None
        if entity_config_str is None:
            return None

        return (entity_config_key, entity_config_str)

    def _search_entity_namespace_in_memory(
        self,
        entity_resource: str,
        grupo_key: str,
        tenant_key: str,
        default_key: str,
    ) -> tuple[str, Namespace] | None:
        namespace = None
        entity_config_key = None

        # Pesquisando a entidade no namespace mais específico (grupo_empresarial)
        grupo_namespace = namespaces_dict.get(grupo_key)
        if grupo_namespace and entity_resource in grupo_namespace.loaded_entities:
            entity_config_key = grupo_key
            namespace = grupo_namespace

        # Pesquisando a entidade no namespace intermediário (tenant)
        tenant_namespace = namespaces_dict.get(tenant_key)
        if tenant_namespace and entity_resource in tenant_namespace.loaded_entities:
            entity_config_key = tenant_key
            namespace = tenant_namespace

        # Pesquisando a entidade no namespace padrão (default)
        default_namespace = namespaces_dict.get(default_key)
        if default_namespace and entity_resource in default_namespace.loaded_entities:
            entity_config_key = default_key
            namespace = default_namespace

        if namespace and entity_config_key:
            return (entity_config_key, namespace)
        else:
            return None



from __future__ import annotations

from typing import Any

from nsj_rest_lib2.compiler.migration_compiler_util import MigrationCompilerUtil


class MigrationCompilerCreateTable:
    def compile(
        self,
        table_name: str,
        column_specs: list[dict[str, Any]],
        pk_columns: list[str],
        fk_specs: list[dict[str, str]],
    ) -> str:
        column_definitions: list[str] = []
        for spec in column_specs:
            col_def = f"{spec['column_name']} {spec['sql_type']}"
            if spec["not_null"]:
                col_def += " NOT NULL"
            if spec["default"] is not None:
                col_def += f" DEFAULT {MigrationCompilerUtil.quote_literal(spec['default'])}"
            column_definitions.append(col_def)

        if pk_columns:
            pk_constraint = (
                f"CONSTRAINT {table_name}_pkey PRIMARY KEY ({', '.join(pk_columns)})"
            )
            column_definitions.append(pk_constraint)

        for fk in fk_specs:
            fk_name = MigrationCompilerUtil.check_constraint_name(
                table_name, fk["column_name"], "fk"
            )
            column_definitions.append(
                f"CONSTRAINT {fk_name} FOREIGN KEY ({fk['column_name']}) REFERENCES {fk['ref_table']}({fk['ref_column']})"
            )

        columns_block = ",\n    ".join(column_definitions)
        create_lines = [
            f"CREATE TABLE {table_name} (",
            f"    {columns_block}",
            ");",
        ]

        for spec in column_specs:
            if spec["description"]:
                create_lines.append(
                    MigrationCompilerUtil.comment_on_column(
                        table_name, spec["column_name"], spec["description"]
                    )
                    + ";"
                )

            if spec["enum_values"]:
                create_lines.extend(
                    [
                        stmt + ";"
                        for stmt in MigrationCompilerUtil.add_enum_check_constraint(
                            table_name, spec["column_name"], spec["enum_values"]
                        )
                    ]
                )

            if spec["maximum"] is not None and spec["is_numeric"]:
                create_lines.extend(
                    [
                        stmt + ";"
                        for stmt in MigrationCompilerUtil.add_max_check_constraint(
                            table_name, spec["column_name"], spec["maximum"]
                        )
                    ]
                )

        return "\n".join(create_lines)



import json
import os
import sys
import copy
from pathlib import Path

import yaml

ROOT_DIR = Path(__file__).resolve().parents[1]
if str(ROOT_DIR) not in sys.path:
    sys.path.insert(0, str(ROOT_DIR))

os.environ.setdefault("ESCOPO_RESTLIB2", "test-scope")

from nsj_rest_lib2.compiler.compiler import EDLCompiler
from nsj_rest_lib2.compiler.edl_model.entity_model import EntityModel
from nsj_rest_lib2.compiler.edl_model.entity_model_root import EntityModelRoot
from nsj_rest_lib2.compiler.migration_compiler import MigrationCompiler
from nsj_rest_lib2.dto.escopo_dto import EscopoDTO


def _load_entity(path: Path) -> tuple[str, EntityModel]:
    with path.open("r") as fp:
        if path.suffix.lower() == ".json":
            edl_data = json.load(fp)
        else:
            edl_data = yaml.safe_load(fp)

    model = (
        EntityModelRoot(**edl_data)
        if edl_data.get("mixin", False)
        else EntityModel(**edl_data)
    )
    complete_id = f"{model.escopo}/{model.id}"
    return complete_id, model


def _load_entity_models(base_dir: Path) -> dict[str, EntityModel]:
    models: dict[str, EntityModel] = {}
    for path in base_dir.iterdir():
        if path.suffix.lower() not in {".json", ".yaml", ".yml"}:
            continue
        key, model = _load_entity(path)
        models[key] = model
    return models


def test_migration_generation_for_pessoa():
    base_dir = Path(__file__).resolve().parents[1] / "@schemas_test"
    entity_models = _load_entity_models(base_dir)

    pessoa_model = entity_models["dados_mestre/pessoa"]

    compiler = EDLCompiler()
    migration_compiler = MigrationCompiler(compiler)

    sql = migration_compiler.compile(pessoa_model, entity_models)

    assert "DO $MIGRATION$" in sql
    assert "IF exists_table('ns.pessoas')" in sql
    assert "CREATE TABLE ns.pessoas" in sql
    assert "pessoa varchar(30)" in sql  # coluna mapeada de 'codigo'
    assert "nome varchar(200)" in sql
    assert "cnpj varchar(20)" in sql  # cpf_cnpj mapeia para coluna cnpj
    assert "ALTER TABLE ns.pessoas ALTER COLUMN nome SET NOT NULL" in sql
    assert "table_has_primary_key('ns.pessoas')" in sql
    # Enum checks gerados
    assert "qualificacao_enum_chk CHECK (qualificacao IN (" in sql
    assert "situacaopagamento_enum_chk CHECK (situacaopagamento IN (" in sql
    assert (
        "indicadorinscricaoestadual_enum_chk CHECK (indicadorinscricaoestadual IN ("
        in sql
    )
    # Comentários de coluna (usa COMMENT ON)
    assert "COMMENT ON COLUMN ns.pessoas" in sql
    # Garantia de uso das funções auxiliares de coluna
    assert "column_default_equals(" in sql
    assert "column_is_not_null(" in sql


def test_migration_generation_with_rename():
    base_dir = Path(__file__).resolve().parents[1] / "@schemas_test"
    entity_models = _load_entity_models(base_dir)

    pessoa_model = entity_models["dados_mestre/pessoa"]
    # Copiando e alterando o mapeamento da coluna 'codigo' de pessoa -> pessoa_new
    import copy

    pessoa_old = copy.deepcopy(pessoa_model)
    pessoa_old.repository.properties["codigo"].column = "pessoa"
    pessoa_new = copy.deepcopy(pessoa_model)
    pessoa_new.repository.properties["codigo"].column = "pessoa_new"

    compiler = EDLCompiler()
    migration_compiler = MigrationCompiler(compiler)

    sql = migration_compiler.compile(pessoa_new, entity_models, pessoa_old)

    assert "RENAME COLUMN pessoa TO pessoa_new" in sql
    assert "exists_column('ns.pessoas', 'pessoa')" in sql
    assert "exists_column('ns.pessoas', 'pessoa_new')" in sql


def test_migration_generation_with_fk():
    base_dir = Path(__file__).resolve().parents[1] / "@schemas_test"
    entity_models = _load_entity_models(base_dir)

    pessoa_model = entity_models["dados_mestre/pessoa"]
    compiler = EDLCompiler()
    migration_compiler = MigrationCompiler(compiler)

    # Simula que 'conta' usa coluna local id_conta apontando para financas/conta
    pessoa_model = copy.deepcopy(pessoa_model)
    pessoa_model.repository.properties["conta"].relation_column = "dados_mestre/pessoa/id_conta"

    sql = migration_compiler.compile(pessoa_model, entity_models)

    assert "id_conta uuid" in sql
    assert "FOREIGN KEY (id_conta)" in sql
    assert "financas.conta" in sql or "financas_conta" in sql


def test_migration_generation_with_remote_fk():
    base_dir = Path(__file__).resolve().parents[1] / "@schemas_test"
    entity_models = _load_entity_models(base_dir)

    endereco_model = entity_models["dados_mestre/endereco"]
    compiler = EDLCompiler()
    migration_compiler = MigrationCompiler(compiler)

    # relation_column de 'enderecos' (em pessoa) aponta coluna id_pessoa na tabela de enderecos
    sql = migration_compiler.compile(endereco_model, entity_models)

    assert "ALTER TABLE ns.enderecos" in sql
    assert "FOREIGN KEY (id_pessoa)" in sql
    assert "REFERENCES ns.pessoas" in sql


def test_compile_external_component_relation():
    base_dir = Path(__file__).resolve().parents[1] / "@schemas_test"
    entity_models = _load_entity_models(base_dir)

    pedido_edl = {
        "edl_version": "1.0",
        "escopo": "vendas",
        "description": "Pedido com relação a componente externa",
        "id": "pedido",
        "version": "1.0",
        "properties": {
            "id": {"type": "uuid", "pk": True},
            "numero": {"type": "string"},
            "contatos": {
                "type": "dados_mestre/pessoa/#/components/contato",
                "label": "Contatos",
                "cardinality": "1_N",
            },
        },
        "repository": {
            "map": "ns.pedidos",
            "table_owner": True,
            "properties": {
                "id": {"column": "id"},
                "numero": {"column": "numero"},
                "contatos": {
                    "relation_column": "dados_mestre/pessoa/#/components/contato/id_pedido"
                },
            },
            "indexes": [],
        },
        "api": {
            "resource": "pedidos",
            "expose": True,
            "verbs": ["GET"],
        },
    }

    pedido_model = EntityModel(**pedido_edl)

    compiler = EDLCompiler()
    escopo = EscopoDTO(codigo=pedido_model.escopo, service_account=None)
    result = compiler.compile_model(
        pedido_model,
        list(entity_models.items()),
        escopo=escopo,
    )

    assert result is not None
    assert "PessoaContatoDTO" in (result.dto_code or "")

    relation_resources = [
        rd.entity_resource for rd in (result.relations_dependencies or [])
    ]
    assert "pessoas" in relation_resources


def test_compile_deep_component_relation():
    pessoa_nested = {
        "edl_version": "1.0",
        "escopo": "dados_mestre",
        "description": "Pessoa com contato e telefone",
        "id": "pessoa",
        "properties": {
            "id": {"type": "uuid", "pk": True},
            "contatos": {
                "type": "#/components/contato",
                "cardinality": "1_N",
                "label": "Contatos",
            },
        },
        "components": {
            "contato": {
                "id": "contato",
                "description": "Contato de pessoa",
                "properties": {
                    "id": {"type": "uuid", "pk": True},
                    "id_pessoa": {"type": "uuid"},
                    "telefones": {
                        "type": "#/components/telefone",
                        "cardinality": "1_N",
                    },
                },
                "components": {
                    "telefone": {
                        "id": "telefone",
                        "description": "Telefone do contato",
                        "properties": {
                            "id": {"type": "uuid", "pk": True},
                            "id_contato": {"type": "uuid"},
                            "numero": {"type": "string"},
                        },
                        "repository": {
                            "map": "ns.telefones",
                            "table_owner": True,
                            "sharedTable": True,
                            "properties": {
                                "id": {"column": "id"},
                                "id_contato": {"column": "id_contato"},
                                "numero": {"column": "numero"},
                            },
                            "indexes": [],
                        },
                    }
                },
                "repository": {
                    "map": "ns.contatos",
                    "table_owner": True,
                    "sharedTable": True,
                    "properties": {
                        "id": {"column": "id"},
                        "id_pessoa": {"column": "id_pessoa"},
                        "telefones": {
                            "relation_column": "dados_mestre/pessoa/#/components/contato/#/components/telefone/id_contato"
                        },
                    },
                    "indexes": [],
                },
            },
        },
        "repository": {
            "map": "ns.pessoas",
            "table_owner": True,
            "properties": {
                "id": {"column": "id"},
                "contatos": {
                    "relation_column": "dados_mestre/pessoa/#/components/contato/id_pessoa"
                },
            },
            "indexes": [],
        },
        "api": {
            "resource": "pessoas",
            "expose": True,
            "verbs": ["GET"],
        },
    }

    followup_edl = {
        "edl_version": "1.0",
        "escopo": "crm",
        "description": "Followup ligado a telefone de contato de pessoa",
        "id": "followup",
        "properties": {
            "id": {"type": "uuid", "pk": True},
            "descricao": {"type": "string"},
            "telefones": {
                "type": "dados_mestre/pessoa/#/components/contato/#/components/telefone",
                "cardinality": "1_N",
            },
        },
        "repository": {
            "map": "ns.followups",
            "table_owner": True,
            "properties": {
                "id": {"column": "id"},
                "descricao": {"column": "descricao"},
                "telefones": {
                    "relation_column": "dados_mestre/pessoa/#/components/contato/#/components/telefone/id_followup"
                },
            },
            "indexes": [],
        },
        "api": {
            "resource": "followups",
            "expose": True,
            "verbs": ["GET"],
        },
    }

    pessoa_model = EntityModel(**pessoa_nested)
    followup_model = EntityModel(**followup_edl)

    entity_models = {
        "dados_mestre/pessoa": pessoa_model,
    }

    compiler = EDLCompiler()
    escopo = EscopoDTO(codigo=followup_model.escopo, service_account=None)
    result = compiler.compile_model(
        followup_model,
        list(entity_models.items()),
        escopo=escopo,
    )

    assert result is not None
    assert "PessoaContatoTelefoneDTO" in (result.dto_code or "")
    assert any(
        rd.entity_resource == "pessoas" for rd in (result.relations_dependencies or [])
    )


def test_compile_deep_component_relation_without_markers():
    pessoa_nested = {
        "edl_version": "1.0",
        "escopo": "dados_mestre",
        "description": "Pessoa com contato e telefone",
        "id": "pessoa",
        "properties": {
            "id": {"type": "uuid", "pk": True},
            "contatos": {
                "type": "#/components/contato",
                "cardinality": "1_N",
                "label": "Contatos",
            },
        },
        "components": {
            "contato": {
                "id": "contato",
                "description": "Contato de pessoa",
                "properties": {
                    "id": {"type": "uuid", "pk": True},
                    "id_pessoa": {"type": "uuid"},
                    "telefones": {
                        "type": "#/components/telefone",
                        "cardinality": "1_N",
                    },
                },
                "components": {
                    "telefone": {
                        "id": "telefone",
                        "description": "Telefone do contato",
                        "properties": {
                            "id": {"type": "uuid", "pk": True},
                            "id_contato": {"type": "uuid"},
                            "numero": {"type": "string"},
                        },
                        "repository": {
                            "map": "ns.telefones",
                            "table_owner": True,
                            "sharedTable": True,
                            "properties": {
                                "id": {"column": "id"},
                                "id_contato": {"column": "id_contato"},
                                "numero": {"column": "numero"},
                            },
                            "indexes": [],
                        },
                    }
                },
                "repository": {
                    "map": "ns.contatos",
                    "table_owner": True,
                    "sharedTable": True,
                    "properties": {
                        "id": {"column": "id"},
                        "id_pessoa": {"column": "id_pessoa"},
                        "telefones": {
                            "relation_column": "dados_mestre/pessoa/contato/telefone/id_contato"
                        },
                    },
                    "indexes": [],
                },
            },
        },
        "repository": {
            "map": "ns.pessoas",
            "table_owner": True,
            "properties": {
                "id": {"column": "id"},
                "contatos": {
                    "relation_column": "dados_mestre/pessoa/contato/id_pessoa"
                },
            },
            "indexes": [],
        },
        "api": {
            "resource": "pessoas",
            "expose": True,
            "verbs": ["GET"],
        },
    }

    followup_edl = {
        "edl_version": "1.0",
        "escopo": "crm",
        "description": "Followup ligado a telefone de contato de pessoa",
        "id": "followup",
        "properties": {
            "id": {"type": "uuid", "pk": True},
            "descricao": {"type": "string"},
            "telefones": {
                "type": "dados_mestre/pessoa/contato/telefone",
                "cardinality": "1_N",
            },
        },
        "repository": {
            "map": "ns.followups",
            "table_owner": True,
            "properties": {
                "id": {"column": "id"},
                "descricao": {"column": "descricao"},
                "telefones": {
                    "relation_column": "dados_mestre/pessoa/contato/telefone/id_followup"
                },
            },
            "indexes": [],
        },
        "api": {
            "resource": "followups",
            "expose": True,
            "verbs": ["GET"],
        },
    }

    pessoa_model = EntityModel(**pessoa_nested)
    followup_model = EntityModel(**followup_edl)

    entity_models = {
        "dados_mestre/pessoa": pessoa_model,
    }

    compiler = EDLCompiler()
    escopo = EscopoDTO(codigo=followup_model.escopo, service_account=None)
    result = compiler.compile_model(
        followup_model,
        list(entity_models.items()),
        escopo=escopo,
    )

    assert result is not None
    assert "PessoaContatoTelefoneDTO" in (result.dto_code or "")
    assert any(
        rd.entity_resource == "pessoas" for rd in (result.relations_dependencies or [])
    )



from __future__ import annotations

import json
from typing import Any, Dict

from nsj_rest_lib2.compiler.edl_model.entity_model import EntityModel
from nsj_rest_lib2.compiler.edl_model.entity_model_base import EntityModelBase
from nsj_rest_lib2.compiler.model import CompilerResult
from nsj_rest_lib2.compiler.util.type_naming_util import (
    build_entity_hash,
    resolve_namespace_key,
)
from nsj_rest_lib2.redis_config import set_redis
from nsj_rest_lib2.settings import ESCOPO_RESTLIB2


class EntityConfigWriter:
    """
    Serializa o resultado da compilação e publica no Redis usando o mesmo
    layout consumido pelo EntityLoader.
    """

    def __init__(self, escopo: str = ESCOPO_RESTLIB2) -> None:
        self._escopo = escopo or ESCOPO_RESTLIB2

    def publish(
        self,
        entity_model: EntityModelBase,
        compiler_result: CompilerResult,
        *,
        entity_hash: str | None = None,
    ) -> Dict[str, Any]:
        """
        Serializa e grava o resultado da compilação no Redis.

        :return: Payload serializado (útil para inspeção/testes).
        """

        payload = self.build_payload(
            entity_model, compiler_result, entity_hash=entity_hash
        )
        namespace_key = self._resolve_namespace_key(entity_model)
        resource = self._resolve_resource(entity_model)

        set_redis(
            "entity_config",
            self._escopo,
            namespace_key,
            resource,
            json.dumps(payload, ensure_ascii=False),
        )

        return payload

    def build_payload(
        self,
        entity_model: EntityModelBase,
        compiler_result: CompilerResult,
        *,
        entity_hash: str | None = None,
    ) -> Dict[str, Any]:
        """
        Constrói o dicionário serializável que representa a entidade compilada.
        """

        resource = self._resolve_resource(entity_model)
        relations = [
            rd.to_dict()
            for rd in (compiler_result.relations_dependencies or [])
        ]

        return {
            "dto_class_name": compiler_result.dto_class_name,
            "entity_class_name": compiler_result.entity_class_name,
            "service_account": compiler_result.service_account,
            "insert_function_class_name": compiler_result.insert_function_class_name,
            "insert_function_name": compiler_result.insert_function_name,
            "source_insert_function": compiler_result.source_insert_function,
            "update_function_class_name": compiler_result.update_function_class_name,
            "update_function_name": compiler_result.update_function_name,
            "source_update_function": compiler_result.source_update_function,
            "get_function_name": compiler_result.get_function_name,
            "list_function_name": compiler_result.list_function_name,
            "delete_function_name": compiler_result.delete_function_name,
            "get_function_type_class_name": compiler_result.get_function_type_class_name,
            "list_function_type_class_name": compiler_result.list_function_type_class_name,
            "delete_function_type_class_name": compiler_result.delete_function_type_class_name,
            "source_get_function_type": compiler_result.source_get_function_type,
            "source_list_function_type": compiler_result.source_list_function_type,
            "source_delete_function_type": compiler_result.source_delete_function_type,
            "retrieve_after_insert": compiler_result.retrieve_after_insert,
            "retrieve_after_update": compiler_result.retrieve_after_update,
            "retrieve_after_partial_update": compiler_result.retrieve_after_partial_update,
            "post_response_dto_class_name": compiler_result.post_response_dto_class_name,
            "put_response_dto_class_name": compiler_result.put_response_dto_class_name,
            "patch_response_dto_class_name": compiler_result.patch_response_dto_class_name,
            "custom_json_post_response": compiler_result.custom_json_post_response,
            "custom_json_put_response": compiler_result.custom_json_put_response,
            "custom_json_patch_response": compiler_result.custom_json_patch_response,
            "custom_json_get_response": compiler_result.custom_json_get_response,
            "custom_json_list_response": compiler_result.custom_json_list_response,
            "custom_json_delete_response": compiler_result.custom_json_delete_response,
            "source_dto": compiler_result.dto_code,
            "source_entity": compiler_result.entity_code,
            "entity_hash": entity_hash
            if entity_hash is not None
            else self._build_entity_hash(compiler_result),
            "api_expose": compiler_result.api_expose,
            "api_resource": resource,
            "api_verbs": compiler_result.api_verbs or [],
            "relations_dependencies": relations,
        }

    def _resolve_namespace_key(self, entity_model: EntityModelBase) -> str:
        tenant = getattr(entity_model, "tenant", None)
        grupo_empresarial = getattr(entity_model, "grupo_empresarial", None)

        return resolve_namespace_key(tenant, grupo_empresarial)

    def _resolve_resource(self, entity_model: EntityModelBase) -> str:
        if not isinstance(entity_model, EntityModel) or not entity_model.api:
            raise ValueError(
                "EntityModel precisa possuir bloco de API para publicação no Redis."
            )
        if not entity_model.api.resource:
            raise ValueError("api.resource é obrigatório para publicação no Redis.")
        return entity_model.api.resource

    def _build_entity_hash(self, compiler_result: CompilerResult) -> str:
        return build_entity_hash(compiler_result)



import logging
import os

ESCOPO_RESTLIB2 = os.environ["ESCOPO_RESTLIB2"]
APP_NAME = os.getenv("APP_NAME", "nsj_rest_lib2")

ENV = os.getenv("ENV", "dev").lower()

if ENV in ("dev", "local"):
    MIN_TIME_SOURCE_REFRESH = 0
else:
    MIN_TIME_SOURCE_REFRESH = int(os.getenv("MIN_TIME_SOURCE_REFRESH", "15"))


def _get_env_bool(name: str, default: bool) -> bool:
    value = os.getenv(name)
    if value is None:
        return default
    return value.strip().lower() in ("1", "true", "t", "yes", "y", "on")


EDLS_FROM_REDIS = _get_env_bool("EDLS_FROM_REDIS", True)


def get_logger():
    return logging.getLogger(APP_NAME)



from __future__ import annotations

from typing import Any

from nsj_rest_lib2.compiler.migration_compiler_util import MigrationCompilerUtil


class MigrationCompilerAlterTable:
    def compile(
        self,
        table_name: str,
        column_specs: list[dict[str, Any]],
        pk_columns: list[str],
        rename_operations: list[tuple[str, str]] | None = None,
        fk_specs: list[dict[str, str]] | None = None,
    ) -> list[str]:
        lines: list[str] = []
        rename_operations = rename_operations or []
        fk_specs = fk_specs or []

        for old_col, new_col in rename_operations:
            lines.append(
                f"        IF exists_column('{table_name}', '{old_col}') AND NOT exists_column('{table_name}', '{new_col}') THEN"
            )
            lines.append(
                f"            ALTER TABLE {table_name} RENAME COLUMN {old_col} TO {new_col};"
            )
            lines.append("        END IF;")

        for spec in column_specs:
            column_name = spec["column_name"]
            default_literal = (
                MigrationCompilerUtil.quote_literal(spec["default"])
                if spec["default"] is not None
                else None
            )
            default_literal_text = (
                f"{default_literal}::text" if default_literal is not None else None
            )
            not_null_literal = "TRUE" if spec["not_null"] else "FALSE"

            lines.append(f"        -- coluna {column_name}")
            lines.append(
                f"        IF NOT exists_column('{table_name}', '{column_name}') THEN"
            )
            lines.append(
                MigrationCompilerUtil.indent_sql(
                    self._build_add_column_sql(table_name, spec), 12
                )[0]
            )

            self._append_enum_and_max_checks(lines, table_name, column_name, spec, 12)

            if spec["description"]:
                lines.append(
                    "            "
                    + MigrationCompilerUtil.comment_on_column(
                        table_name, column_name, spec["description"]
                    )
                    + ";"
                )

            lines.append("        ELSE")

            # Alteração de tipo
            lines.append(
                f"            IF NOT validate_column_properties('{table_name}', '{column_name}', '{spec['sql_type']}') THEN"
            )
            lines.append(
                f"                ALTER TABLE {table_name} ALTER COLUMN {column_name} TYPE {spec['sql_type']};"
            )
            lines.append("            END IF;")

            # Alteração de nulidade
            lines.append(f"            IF {not_null_literal} THEN")
            lines.append(
                f"                IF NOT column_is_not_null('{table_name}', '{column_name}') THEN"
            )
            lines.append(
                f"                    ALTER TABLE {table_name} ALTER COLUMN {column_name} SET NOT NULL;"
            )
            lines.append("                END IF;")
            lines.append("            ELSE")
            lines.append(
                f"                IF column_is_not_null('{table_name}', '{column_name}') THEN"
            )
            lines.append(
                f"                    ALTER TABLE {table_name} ALTER COLUMN {column_name} DROP NOT NULL;"
            )
            lines.append("                END IF;")
            lines.append("            END IF;")

            # Alteração de default
            if default_literal is None:
                lines.append(
                    f"            IF column_default_expr('{table_name}', '{column_name}') IS NOT NULL THEN"
                )
                lines.append(
                    f"                ALTER TABLE {table_name} ALTER COLUMN {column_name} DROP DEFAULT;"
                )
                lines.append("            END IF;")
            else:
                lines.append(
                    f"            IF NOT column_default_equals('{table_name}', '{column_name}', {default_literal_text}) THEN"
                )
                lines.append(
                    f"                ALTER TABLE {table_name} ALTER COLUMN {column_name} SET DEFAULT {default_literal};"
                )
                lines.append("            END IF;")

            # Checks e comentários
            self._append_enum_and_max_checks(lines, table_name, column_name, spec, 12)

            if spec["description"]:
                lines.append(
                    "            "
                    + MigrationCompilerUtil.comment_on_column(
                        table_name, column_name, spec["description"]
                    )
                    + ";"
                )

            lines.append("        END IF;")

        if pk_columns:
            lines.append(f"        IF NOT table_has_primary_key('{table_name}') THEN")
            lines.append(
                f"            {MigrationCompilerUtil.add_primary_key(table_name, pk_columns)};"
            )
            lines.append("        END IF;")

        for fk in fk_specs:
            lines.append(
                f"        IF NOT fk_constraint_matches('{table_name}', '{fk['column_name']}', '{fk['ref_table']}', '{fk['ref_column']}') THEN"
            )
            fk_constraint_name = MigrationCompilerUtil.check_constraint_name(
                table_name, fk["column_name"], "fk"
            )
            lines.append(
                f"            {MigrationCompilerUtil.drop_constraint_if_exists(table_name, fk_constraint_name)};"
            )
            lines.append(
                f"            ALTER TABLE {table_name} ADD CONSTRAINT {fk_constraint_name} FOREIGN KEY ({fk['column_name']}) REFERENCES {fk['ref_table']}({fk['ref_column']});"
            )
            lines.append("        END IF;")

        return lines

    def _append_enum_and_max_checks(
        self,
        lines: list[str],
        table_name: str,
        column_name: str,
        spec: dict[str, Any],
        indent: int,
    ) -> None:
        if spec["enum_values"]:
            formatted_values = ", ".join(
                MigrationCompilerUtil.quote_literal(v) for v in spec["enum_values"]
            )
            lines.append(
                f"{' ' * indent}IF NOT enum_constraint_matches('{table_name}', '{column_name}', ARRAY[{formatted_values}]) THEN"
            )
            lines.extend(
                MigrationCompilerUtil.indent_sql(
                    ";\n".join(
                        MigrationCompilerUtil.add_enum_check_constraint(
                            table_name, column_name, spec["enum_values"]
                        )
                    )
                    + ";",
                    indent + 4,
                )
            )
            lines.append(f"{' ' * indent}END IF;")

        if spec["maximum"] is not None and spec["is_numeric"]:
            lines.append(
                f"{' ' * indent}IF NOT max_constraint_matches('{table_name}', '{column_name}', {spec['maximum']}) THEN"
            )
            lines.extend(
                MigrationCompilerUtil.indent_sql(
                    ";\n".join(
                        MigrationCompilerUtil.add_max_check_constraint(
                            table_name, column_name, spec["maximum"]
                        )
                    )
                    + ";",
                    indent + 4,
                )
            )
            lines.append(f"{' ' * indent}END IF;")

    def _build_add_column_sql(self, table_name: str, spec: dict[str, Any]) -> str:
        clause = f"ALTER TABLE {table_name} ADD COLUMN {spec['column_name']} {spec['sql_type']}"
        if spec["not_null"]:
            clause += " NOT NULL"
        if spec["default"] is not None:
            clause += f" DEFAULT {MigrationCompilerUtil.quote_literal(spec['default'])}"
        clause += ";"
        return clause



from __future__ import annotations

import ast
import re
from dataclasses import dataclass, field
from typing import Dict, Optional, Sequence

import black

from nsj_rest_lib2.compiler.compiler_structures import PropertiesCompilerStructure
from nsj_rest_lib2.compiler.edl_model.api_model import HandlerConfig, HandlerMapping
from nsj_rest_lib2.compiler.edl_model.entity_model_base import EntityModelBase
from nsj_rest_lib2.compiler.edl_model.primitives import PrimitiveTypes
from nsj_rest_lib2.compiler.function_model import (
    FunctionBindingConfig,
    FunctionCompilationOutput,
    FunctionRelationBinding,
)
from nsj_rest_lib2.compiler.util.str_util import CompilerStrUtil
from nsj_rest_lib2.compiler.util.type_naming_util import compile_function_class_name
from nsj_rest_lib2.compiler.util.type_util import TypeUtil


@dataclass
class _MappingNode:
    attr: str
    python_name: str
    dto_field: Optional[str]
    children: list["_MappingNode"] = field(default_factory=list)
    pg_type: Optional[str] = None
    is_array: bool = False


@dataclass
class _FieldSpec:
    name: str
    annotation: str
    type_field_name: Optional[str]
    relation_class: Optional[str]
    is_relation: bool
    is_list: bool


@dataclass
class _ClassSpec:
    name: str
    type_name: str
    function_name: str
    fields: list[_FieldSpec]
    children: list["_ClassSpec"] = field(default_factory=list)


class FunctionInsertUpdateCompiler:
    """
    Responsável por compilar tipos usados por funções de banco (Insert/Update)
    com base na configuração declarada no EDL.
    """

    def compile_insert(
        self,
        entity_model: EntityModelBase,
        properties_structure: PropertiesCompilerStructure,
        handler_config: Optional[HandlerConfig],
        prefx_class_name: str,
    ) -> FunctionCompilationOutput:
        return self._compile(
            entity_model,
            properties_structure,
            handler_config,
            prefx_class_name,
            operation="insert",
        )

    def compile_update(
        self,
        entity_model: EntityModelBase,
        properties_structure: PropertiesCompilerStructure,
        handler_config: Optional[HandlerConfig],
        prefx_class_name: str,
    ) -> FunctionCompilationOutput:
        return self._compile(
            entity_model,
            properties_structure,
            handler_config,
            prefx_class_name,
            operation="update",
        )

    # ------------------------------------------------------------------ #
    # Internal helpers
    # ------------------------------------------------------------------ #

    def _compile(
        self,
        entity_model: EntityModelBase,
        properties_structure: PropertiesCompilerStructure,
        handler_config: Optional[HandlerConfig],
        prefx_class_name: str,
        operation: str,
    ) -> FunctionCompilationOutput:
        if not handler_config or handler_config.impl != "pg_function":
            return FunctionCompilationOutput()

        if not handler_config.call or not handler_config.call.arg_binding:
            return FunctionCompilationOutput()

        arg_binding = handler_config.call.arg_binding
        if not arg_binding.mapping or not arg_binding.type_name:
            return FunctionCompilationOutput()

        function_name = handler_config.function_ref
        if not function_name:
            return FunctionCompilationOutput()

        mapping_nodes = self._build_nodes(arg_binding.mapping)
        if not mapping_nodes:
            return FunctionCompilationOutput()

        class_name = compile_function_class_name(
            entity_model.id,
            prefx_class_name,
            [],
            operation,
        )

        field_bindings: Dict[str, str] = {}
        relation_bindings: Dict[str, FunctionRelationBinding] = {}

        class_spec = self._build_class_spec(
            mapping_nodes,
            entity_model,
            properties_structure,
            function_name,
            class_name,
            arg_binding.type_name,
            path=[],
            operation=operation,
            field_bindings=field_bindings,
            relation_bindings=relation_bindings,
            prefx_class_name=prefx_class_name,
        )

        module_ast, has_relations, needs_any = self._build_module_ast(
            class_spec, operation
        )

        if not module_ast:
            return FunctionCompilationOutput()

        imports = self._build_imports(operation, has_relations, needs_any)
        module = ast.Module(body=imports + module_ast, type_ignores=[])
        module = ast.fix_missing_locations(module)
        code = ast.unparse(module)
        code = black.format_str(code, mode=black.FileMode())

        return FunctionCompilationOutput(
            class_name=class_name,
            code=code,
            function_name=function_name,
            field_bindings=field_bindings,
            relation_bindings=relation_bindings,
        )

    def _build_nodes(self, mappings: Sequence[HandlerMapping]) -> list[_MappingNode]:
        nodes: list[_MappingNode] = []
        used_names: Dict[str, int] = {}

        for entry in mappings or []:
            if not entry.attr:
                continue

            python_name = self._sanitize_identifier(entry.attr)
            python_name = self._ensure_unique_name(python_name, used_names)
            dto_field = self._extract_dto_field(entry.from_)

            node = _MappingNode(
                attr=entry.attr,
                python_name=python_name,
                dto_field=dto_field,
            )

            if entry.mapping:
                pg_type, is_array = self._parse_type_declaration(entry.as_)
                node.pg_type = pg_type
                node.is_array = is_array
                node.children = self._build_nodes(entry.mapping)

            nodes.append(node)

        return nodes

    def _build_class_spec(
        self,
        nodes: Sequence[_MappingNode],
        entity_model: EntityModelBase,
        properties_structure: PropertiesCompilerStructure,
        function_name: str,
        class_name: str,
        type_name: str,
        path: list[str],
        operation: str,
        field_bindings: Dict[str, str],
        relation_bindings: Dict[str, FunctionRelationBinding],
        prefx_class_name: str,
    ) -> _ClassSpec:
        fields: list[_FieldSpec] = []
        children_specs: list[_ClassSpec] = []

        for node in nodes:
            if node.children:
                if not node.pg_type:
                    raise ValueError(
                        f"O campo '{node.attr}' possui sub-mapeamento, porém o tipo Postgres não foi informado em 'as'."
                    )

                child_class_name = compile_function_class_name(
                    entity_model.id,
                    prefx_class_name,
                    path + [node.python_name],
                    operation,
                )

                child_spec = self._build_class_spec(
                    node.children,
                    entity_model,
                    properties_structure,
                    function_name,
                    child_class_name,
                    node.pg_type,
                    path + [node.python_name],
                    operation,
                    field_bindings,
                    relation_bindings,
                    prefx_class_name,
                )

                children_specs.append(child_spec)
                fields.append(
                    _FieldSpec(
                        name=node.python_name,
                        annotation=self._relation_annotation(
                            child_class_name, node.is_array
                        ),
                        type_field_name=(
                            node.attr if node.python_name != node.attr else None
                        ),
                        relation_class=child_class_name,
                        is_relation=True,
                        is_list=node.is_array,
                    )
                )

                if node.dto_field:
                    relation_bindings[node.dto_field] = FunctionRelationBinding(
                        field_name=node.python_name,
                        function_type_class=child_class_name,
                    )

                continue

            annotation = self._resolve_annotation(node, properties_structure)
            fields.append(
                _FieldSpec(
                    name=node.python_name,
                    annotation=annotation,
                    type_field_name=(
                        node.attr if node.python_name != node.attr else None
                    ),
                    relation_class=None,
                    is_relation=False,
                    is_list=False,
                )
            )

            if node.dto_field and node.dto_field != node.python_name:
                field_bindings[node.dto_field] = node.python_name

        return _ClassSpec(
            name=class_name,
            type_name=type_name,
            function_name=function_name,
            fields=fields,
            children=children_specs,
        )

    def _build_module_ast(
        self,
        spec: _ClassSpec,
        operation: str,
    ) -> tuple[list[ast.stmt], bool, bool]:
        class_defs: list[ast.stmt] = []
        has_relation = False
        needs_any = False

        for child in spec.children:
            child_defs, child_has_relation, child_needs_any = self._build_module_ast(
                child, operation
            )
            class_defs.extend(child_defs)
            has_relation = has_relation or child_has_relation
            needs_any = needs_any or child_needs_any

        body: list[ast.stmt] = []
        for field in spec.fields:
            annotation_expr = self._build_annotation_expr(field.annotation, field)
            descriptor_call = self._build_descriptor_call(field)
            body.append(
                ast.AnnAssign(
                    target=ast.Name(id=field.name, ctx=ast.Store()),
                    annotation=annotation_expr,
                    value=descriptor_call,
                    simple=1,
                )
            )
            if field.is_relation:
                has_relation = True
            if field.annotation == "Any":
                needs_any = True

        if not body:
            body = [ast.Pass()]

        decorator_name = (
            "InsertFunctionType" if operation == "insert" else "UpdateFunctionType"
        )
        base_class = (
            "InsertFunctionTypeBase"
            if operation == "insert"
            else "UpdateFunctionTypeBase"
        )

        decorator = ast.Call(
            func=ast.Name(id=decorator_name, ctx=ast.Load()),
            args=[],
            keywords=[
                ast.keyword(arg="type_name", value=ast.Constant(value=spec.type_name)),
            ],
        )

        class_defs.append(
            ast.ClassDef(
                name=spec.name,
                bases=[ast.Name(id=base_class, ctx=ast.Load())],
                keywords=[],
                decorator_list=[decorator],
                body=body,
            )
        )

        return class_defs, has_relation, needs_any

    def _build_imports(
        self, operation: str, has_relation_fields: bool, needs_any: bool
    ) -> list[ast.stmt]:
        imports: list[ast.stmt] = [
            ast.Import(names=[ast.alias(name="datetime", asname=None)]),
            ast.Import(names=[ast.alias(name="uuid", asname=None)]),
        ]

        if needs_any:
            imports.append(
                ast.ImportFrom(
                    module="typing",
                    names=[ast.alias(name="Any", asname=None)],
                    level=0,
                )
            )

        decorator_module = (
            "nsj_rest_lib.decorator.insert_function_type"
            if operation == "insert"
            else "nsj_rest_lib.decorator.update_function_type"
        )
        decorator_name = (
            "InsertFunctionType" if operation == "insert" else "UpdateFunctionType"
        )

        base_module = "nsj_rest_lib.entity.function_type_base"
        base_name = (
            "InsertFunctionTypeBase"
            if operation == "insert"
            else "UpdateFunctionTypeBase"
        )

        imports.extend(
            [
                ast.ImportFrom(
                    module=decorator_module,
                    names=[ast.alias(name=decorator_name, asname=None)],
                    level=0,
                ),
                ast.ImportFrom(
                    module="nsj_rest_lib.descriptor.function_field",
                    names=[ast.alias(name="FunctionField", asname=None)],
                    level=0,
                ),
                ast.ImportFrom(
                    module=base_module,
                    names=[ast.alias(name=base_name, asname=None)],
                    level=0,
                ),
            ]
        )

        if has_relation_fields:
            imports.append(
                ast.ImportFrom(
                    module="nsj_rest_lib.descriptor.function_relation_field",
                    names=[ast.alias(name="FunctionRelationField", asname=None)],
                    level=0,
                )
            )

        return imports

    def _build_annotation_expr(
        self, annotation: str, field: _FieldSpec
    ) -> ast.expr:
        if field.is_relation and field.is_list:
            return ast.Subscript(
                value=ast.Name(id="list", ctx=ast.Load()),
                slice=ast.Name(id=field.relation_class, ctx=ast.Load()),
                ctx=ast.Load(),
            )

        if field.is_relation:
            return ast.Name(id=field.relation_class, ctx=ast.Load())

        if annotation == "Any":
            return ast.Name(id="Any", ctx=ast.Load())

        if "." in annotation:
            parts = annotation.split(".")
            expr = ast.Name(id=parts[0], ctx=ast.Load())
            for part in parts[1:]:
                expr = ast.Attribute(value=expr, attr=part, ctx=ast.Load())
            return expr

        return ast.Name(id=annotation, ctx=ast.Load())

    def _build_descriptor_call(self, field: _FieldSpec) -> ast.Call:
        descriptor_cls = (
            "FunctionRelationField" if field.is_relation else "FunctionField"
        )
        keywords: list[ast.keyword] = []
        if field.type_field_name:
            keywords.append(
                ast.keyword(
                    arg="type_field_name",
                    value=ast.Constant(value=field.type_field_name),
                )
            )
        return ast.Call(
            func=ast.Name(id=descriptor_cls, ctx=ast.Load()),
            args=[],
            keywords=keywords,
        )

    def _resolve_annotation(
        self,
        node: _MappingNode,
        properties_structure: PropertiesCompilerStructure,
    ) -> str:
        dto_field = node.dto_field
        if not dto_field:
            return "Any"

        prop = properties_structure.properties.get(dto_field) if properties_structure else None
        if not prop or not isinstance(prop.type, PrimitiveTypes):
            return "Any"

        return TypeUtil.property_type_to_python_type(prop.type)

    # ------------------------------------------------------------------ #
    # Utility helpers
    # ------------------------------------------------------------------ #

    def _sanitize_identifier(self, name: str) -> str:
        candidate = CompilerStrUtil.to_snake_case(name)
        candidate = re.sub(r"\W+", "_", candidate)
        if not candidate:
            candidate = "field"
        if candidate[0].isdigit():
            candidate = f"_{candidate}"
        return candidate

    def _ensure_unique_name(self, name: str, used: Dict[str, int]) -> str:
        counter = used.setdefault(name, 0)
        if counter == 0:
            used[name] = 1
            return name

        while True:
            candidate = f"{name}_{counter}"
            counter += 1
            if candidate not in used:
                used[name] = counter
                used[candidate] = 1
                return candidate

    def _extract_dto_field(self, source: Optional[str]) -> Optional[str]:
        if not source:
            return None

        if source.startswith("body."):
            segment = source[5:].split(".", 1)[0]
            return CompilerStrUtil.to_snake_case(segment)

        return None

    def _parse_type_declaration(self, declaration: Optional[str]) -> tuple[str, bool]:
        if not declaration:
            return (None, False)

        declaration = declaration.strip()
        if declaration.endswith("[]"):
            return (declaration[:-2], True)
        return (declaration, False)

    def _relation_annotation(self, class_name: str, is_array: bool) -> str:
        if is_array:
            return f"list[{class_name}]"
        return class_name


def inject_function_bindings(
    target_config: FunctionBindingConfig,
    insert_output: FunctionCompilationOutput,
    update_output: FunctionCompilationOutput,
) -> FunctionBindingConfig:
    """
    Popula um FunctionBindingConfig a partir dos resultados de compilação das
    funções de insert/update.
    """

    for dto_field, function_field in insert_output.field_bindings.items():
        target_config.insert_fields[dto_field] = function_field

    for dto_field, relation in insert_output.relation_bindings.items():
        target_config.insert_relations[dto_field] = relation

    for dto_field, function_field in update_output.field_bindings.items():
        target_config.update_fields[dto_field] = function_field

    for dto_field, relation in update_output.relation_bindings.items():
        target_config.update_relations[dto_field] = relation

    return target_config



import hashlib
import uuid
from typing import Iterable, Iterator, Optional

from nsj_rest_lib2.compiler.model import CompilerResult
from nsj_rest_lib2.compiler.util.str_util import CompilerStrUtil


def compile_namespace_keys(
    tenant: str | int | None, grupo_empresarial: str | uuid.UUID | None
) -> tuple[str, str, str]:
    grupo_key = f"tenant_{tenant}.ge_{grupo_empresarial}"
    tenant_key = f"tenant_{tenant}"
    default_key = "default"

    return (grupo_key, tenant_key, default_key)


def resolve_namespace_key(
    tenant: str | int | None, grupo_empresarial: str | uuid.UUID | None
) -> str:
    grupo_key, tenant_key, default_key = compile_namespace_keys(
        tenant,
        grupo_empresarial,
    )

    if (
        tenant
        and tenant != 0
        and grupo_empresarial
        and str(grupo_empresarial) != "00000000-0000-0000-0000-000000000000"
    ):
        return grupo_key
    if tenant and tenant != 0:
        return tenant_key
    return default_key


def build_entity_hash(compiler_result: CompilerResult) -> str:
    hasher = hashlib.sha256()

    for content in _iter_hash_chunks(
        compiler_result.dto_code,
        compiler_result.entity_code,
        compiler_result.insert_function_name,
        compiler_result.update_function_name,
        compiler_result.get_function_name,
        compiler_result.list_function_name,
        compiler_result.delete_function_name,
        compiler_result.source_insert_function,
        compiler_result.source_update_function,
        compiler_result.source_get_function_type,
        compiler_result.source_list_function_type,
        compiler_result.source_delete_function_type,
        compiler_result.post_response_dto_class_name,
        compiler_result.put_response_dto_class_name,
        compiler_result.patch_response_dto_class_name,
        str(compiler_result.retrieve_after_insert),
        str(compiler_result.retrieve_after_update),
        str(compiler_result.retrieve_after_partial_update),
        str(compiler_result.custom_json_post_response),
        str(compiler_result.custom_json_put_response),
        str(compiler_result.custom_json_patch_response),
        str(compiler_result.custom_json_get_response),
        str(compiler_result.custom_json_list_response),
        str(compiler_result.custom_json_delete_response),
        compiler_result.service_account,
    ):
        hasher.update(content)

    return hasher.hexdigest()


def _iter_hash_chunks(*chunks: Optional[str]) -> Iterator[bytes]:
    for chunk in chunks:
        if chunk:
            yield chunk.encode("utf-8")


def compile_dto_class_name(entity_id: str, prefx_class_name: str = "") -> str:
    return f"{CompilerStrUtil.to_pascal_case(prefx_class_name)}{CompilerStrUtil.to_pascal_case(entity_id)}DTO"


def compile_response_dto_class_name(
    entity_id: str, prefx_class_name: str, verb: str
) -> str:
    verb = (verb or "").lower()
    suffix_map = {
        "post": "PostResponseDTO",
        "put": "PutResponseDTO",
        "patch": "PatchResponseDTO",
    }
    suffix = suffix_map.get(
        verb, f"{CompilerStrUtil.to_pascal_case(verb)}ResponseDTO"
    )
    return (
        f"{CompilerStrUtil.to_pascal_case(prefx_class_name)}"
        f"{CompilerStrUtil.to_pascal_case(entity_id)}"
        f"{suffix}"
    )


def compile_entity_class_name(entity_id: str, prefx_class_name: str = "") -> str:
    return f"{CompilerStrUtil.to_pascal_case(prefx_class_name)}{CompilerStrUtil.to_pascal_case(entity_id)}Entity"


def compile_function_class_name(
    entity_id: str,
    prefx_class_name: str,
    path_parts: Iterable[str],
    operation: str,
) -> str:
    """
    Gera o nome da classe para FunctionTypes (insert/update),
    incorporando o caminho dos relacionamentos (quando houver).
    """

    operation = (operation or "").lower()
    suffix_map = {
        "insert": "InsertType",
        "update": "UpdateType",
        "get": "GetType",
        "list": "ListType",
        "delete": "DeleteType",
    }
    suffix = suffix_map.get(
        operation, f"{CompilerStrUtil.to_pascal_case(operation)}Type"
    )
    path_suffix = "".join(CompilerStrUtil.to_pascal_case(part) for part in path_parts)
    return (
        f"{CompilerStrUtil.to_pascal_case(prefx_class_name)}"
        f"{CompilerStrUtil.to_pascal_case(entity_id)}"
        f"{path_suffix}"
        f"{suffix}"
    )


def compile_function_params_dto_class_name(
    entity_id: str,
    prefx_class_name: str,
    verb: str,
) -> str:
    """
    Gera o nome da classe DTO usada como envelope de parâmetros para funções
    de GET, LIST e DELETE.
    """

    verb = (verb or "").lower()
    if verb == "get":
        suffix = "GetParamsDTO"
    elif verb == "list":
        suffix = "ListParamsDTO"
    elif verb == "delete":
        suffix = "DeleteParamsDTO"
    else:
        suffix = f"{CompilerStrUtil.to_pascal_case(verb)}ParamsDTO"

    return (
        f"{CompilerStrUtil.to_pascal_case(prefx_class_name)}"
        f"{CompilerStrUtil.to_pascal_case(entity_id)}"
        f"{suffix}"
    )



from dataclasses import dataclass, field
from typing import Any, Optional

from nsj_rest_lib2.compiler.compiler import EDLCompiler
from nsj_rest_lib2.compiler.compiler_structures import PropertiesCompilerStructure
from nsj_rest_lib2.compiler.edl_model.entity_model import EntityModel
from nsj_rest_lib2.compiler.edl_model.entity_model_base import EntityModelBase
from nsj_rest_lib2.compiler.edl_model.primitives import PrimitiveTypes, PropertyType
from nsj_rest_lib2.compiler.migration_compiler_util import MigrationCompilerUtil
from nsj_rest_lib2.compiler.migration_compiler_create_table import (
    MigrationCompilerCreateTable,
)
from nsj_rest_lib2.compiler.migration_compiler_alter_table import (
    MigrationCompilerAlterTable,
)
from nsj_rest_lib2.compiler.migration_compiler_create_table import (
    MigrationCompilerCreateTable,
)
from nsj_rest_lib2.compiler.migration_compiler_alter_table import (
    MigrationCompilerAlterTable,
)
from nsj_rest_lib2.compiler.util.relation_ref import RelationRef, RelationRefParser


class DiffAction:
    ADD_COLUMN = "add_column"
    ALTER_COLUMN = "alter_column"
    DROP_COLUMN = "drop_column"


# TODO Tratar criação e remoção de FKs


@dataclass
class DiffProperty:
    action: str
    property_name: str
    entity_name_old: Optional[str] = None
    entity_name: Optional[str] = None
    old_datatype: Optional[PropertyType] = None
    new_datatype: Optional[PropertyType] = None
    old_required: Optional[bool] = None
    new_required: Optional[bool] = None
    new_description: Optional[str] = None
    new_default: Any = None
    old_pk: Optional[bool] = None
    new_pk: Optional[bool] = None
    new_max_length: Optional[int] = None
    new_check: list[Any] = field(default_factory=list)
    renamed: bool = False
    datatype_changed: bool = False
    required_changed: bool = False
    default_changed: bool = False
    pk_changed: bool = False
    max_length_changed: bool = False
    check_changed: bool = False


class MigrationCompiler:
    def __init__(self, compiler: EDLCompiler):
        self._compiler = compiler
        self._create_table_compiler = MigrationCompilerCreateTable()
        self._alter_table_compiler = MigrationCompilerAlterTable()

    def compile(
        self,
        entity_model: EntityModelBase,
        entity_models: dict[str, EntityModel],
        entity_model_old: EntityModelBase | None = None,
    ) -> str:
        # Resolvendo a estrutura de propriedades da entidade nova
        properties_structure = PropertiesCompilerStructure()
        self._compiler._make_properties_structures(
            properties_structure, entity_model, entity_models
        )
        self._filter_properties_for_migration(properties_structure)

        # Resolvendo a estrutura de propriedades da versão antiga da mesma entidade (para detecção de rename)
        properties_structure_old = None
        if entity_model_old:
            properties_structure_old = PropertiesCompilerStructure()
            self._compiler._make_properties_structures(
                properties_structure_old, entity_model_old, entity_models
            )
            self._filter_properties_for_migration(properties_structure_old)

        table_name = entity_model.repository.map
        column_specs, fk_specs = self._build_column_and_fk_specs(
            entity_model, properties_structure, entity_models
        )
        rename_operations = (
            self._detect_renamed_columns(properties_structure, properties_structure_old)
            if properties_structure_old
            else []
        )
        pk_columns = [spec["column_name"] for spec in column_specs if spec["is_pk"]]

        create_table_sql = self._create_table_compiler.compile(
            table_name, column_specs, pk_columns, fk_specs
        )
        alter_block_sql = self._alter_table_compiler.compile(
            table_name, column_specs, pk_columns, rename_operations, fk_specs
        )

        block_lines = [
            "DO $MIGRATION$",
            "BEGIN",
            f"    IF exists_table('{table_name}') THEN",
        ]
        block_lines.extend(alter_block_sql)
        block_lines.append("    ELSE")
        block_lines.extend(MigrationCompilerUtil.indent_sql(create_table_sql, 8))
        block_lines.append("    END IF;")
        block_lines.append("END$MIGRATION$;")

        return "\n".join(block_lines)

    def _filter_properties_for_migration(
        self, properties_structure: PropertiesCompilerStructure
    ) -> None:
        """
        Remove propriedades herdadas via trait_from ou extends, pois as migrações
        dessas entidades devem ser conduzidas nos EDLs originais.
        """
        allowed_properties: dict[str, Any] = {}
        origins = getattr(properties_structure, "property_origins", {}) or {}
        for name, prop in properties_structure.properties.items():
            origin = origins.get(name, "self")
            if origin in {"trait", "extends"}:
                continue
            allowed_properties[name] = prop

        # Atualiza propriedades e coleções dependentes
        properties_structure.properties = allowed_properties
        allowed_names = set(allowed_properties.keys())
        properties_structure.required = [
            r for r in properties_structure.required if r in allowed_names
        ]
        properties_structure.partition_data = [
            p for p in properties_structure.partition_data if p in allowed_names
        ]
        properties_structure.search_properties = [
            s for s in properties_structure.search_properties if s in allowed_names
        ]
        properties_structure.metric_label = [
            m for m in properties_structure.metric_label if m in allowed_names
        ]
        properties_structure.entity_properties = {
            k: v
            for k, v in properties_structure.entity_properties.items()
            if k in allowed_names
        }

    # FIXME Comentado por hora (tentando confiar mais no BD do que nos EDLs)
    # def _compare_properties_structures(
    #     self,
    #     properties_structure: PropertiesCompilerStructure,
    #     properties_structure_old: PropertiesCompilerStructure,
    # ) -> tuple[list[DiffProperty], list[DiffProperty], list[DiffProperty]]:
    #     alter_properties: list[DiffProperty] = []
    #     add_properties: list[DiffProperty] = []
    #     drop_properties: list[DiffProperty] = []

    #     mapa_equivalencias_old_new: dict[str, str] = {}

    #     # Identificando as propriedades a serem alteradas
    #     for property_name_new in properties_structure.properties:
    #         # Guardando as propriedades da nova coluna
    #         property_new = properties_structure.properties[property_name_new]
    #         entity_mapping_new = properties_structure_old.entity_properties.get(
    #             property_name_new
    #         )
    #         entity_name_new = property_name_new
    #         if entity_mapping_new and entity_mapping_new.column:
    #             entity_name_new = entity_mapping_new.column

    #         # Tentando encontrar uma propriedade equivalente na versão antiga
    #         # É equivalente se tiver o mesmo nome de propriedade, ou se tiver o mesmo nome de coluna

    #         # Recuperando a coluna nova, se houver
    #         achou_property_equivalente = False
    #         property_name_old = None
    #         property_old = None
    #         entity_name_old = None

    #         if (
    #             properties_structure_old
    #             and property_name_new in properties_structure_old.properties
    #         ):
    #             achou_property_equivalente = True

    #             # Pelo nome da propriedade
    #             property_name_old = property_name_new
    #             property_old = properties_structure_old.properties[property_name_old]
    #             entity_mapping_old = properties_structure_old.entity_properties.get(
    #                 property_name_old
    #             )
    #             entity_name_old = property_name_old
    #             if entity_mapping_old and entity_mapping_old.column:
    #                 entity_name_old = entity_mapping_old.column

    #             # Guardando a equivalência para ajudar no drop
    #             mapa_equivalencias_old_new[property_name_old] = property_name_new

    #         elif properties_structure_old:
    #             # Buscando pelo nome da coluna
    #             for property_name_old in properties_structure_old.properties:
    #                 property_old = properties_structure_old.properties[
    #                     property_name_old
    #                 ]
    #                 entity_mapping_old = properties_structure_old.entity_properties.get(
    #                     property_name_old
    #                 )
    #                 entity_name_old = property_name_old
    #                 if entity_mapping_old and entity_mapping_old.column:
    #                     entity_name_old = entity_mapping_old.column

    #                 if entity_name_new == entity_name_old:
    #                     achou_property_equivalente = True

    #                     # Guardando a equivalência para ajudar no drop
    #                     mapa_equivalencias_old_new[property_name_old] = (
    #                         property_name_new
    #                     )

    #                     break

    #         if not achou_property_equivalente:
    #             # Construindo um objeto de adição de coluna
    #             mapped_values_new = []
    #             if property_new.domain_config:
    #                 mapped_values_new = [
    #                     v.mapped_value for v in property_new.domain_config
    #                 ]
    #                 mapped_values_new.sort(key=lambda x: x)

    #             add_properties.append(
    #                 DiffProperty(
    #                     action=DiffAction.ADD_COLUMN,
    #                     property_name=property_name_new,
    #                     entity_name=entity_name_new,
    #                     new_datatype=property_new.type,
    #                     new_required=property_name_new in properties_structure.required,
    #                     new_description=property_new.description,
    #                     new_default=property_new.default,
    #                     new_pk=property_new.pk,
    #                     new_max_length=property_new.max_length,
    #                     new_check=mapped_values_new,
    #                     datatype_changed=True,
    #                     required_changed=property_name_new
    #                     in properties_structure.required,
    #                     default_changed=property_new.default is not None,
    #                     pk_changed=bool(property_new.pk),
    #                     max_length_changed=property_new.max_length is not None,
    #                     check_changed=bool(mapped_values_new),
    #                 )
    #             )

    #         else:

    #             # Construindo um objeto de alteração (para ser usado, se necessário)
    #             persist_alter = False
    #             diff_property = DiffProperty(
    #                 action=DiffAction.ALTER_COLUMN,
    #                 property_name=property_name_new,
    #                 entity_name_old=entity_name_old,
    #                 entity_name=entity_name_new,
    #             )

    #             # Verificando se é alteração do nome da coluna
    #             if entity_name_old != entity_name_new:
    #                 persist_alter = True
    #                 diff_property.entity_name_old = entity_name_old
    #                 diff_property.renamed = True

    #             # Verificando se é alteração de tipo
    #             if property_old.type != property_new.type:
    #                 persist_alter = True
    #                 diff_property.old_datatype = property_old.type
    #                 diff_property.new_datatype = property_new.type
    #                 diff_property.datatype_changed = True

    #             # Verificando se houve alteração de nulidade
    #             required_new = property_name_new in properties_structure.required
    #             required_old = property_name_old in properties_structure_old.required
    #             if required_new != required_old:
    #                 persist_alter = True
    #                 diff_property.old_required = required_old
    #                 diff_property.new_required = required_new
    #                 diff_property.required_changed = True

    #             # Verificando se houve alteração de descrição
    #             if property_old.description != property_new.description:
    #                 persist_alter = True
    #                 diff_property.new_description = property_new.description

    #             # Verificando se houve alteração de valor default
    #             if property_old.default != property_new.default:
    #                 persist_alter = True
    #                 diff_property.new_default = property_new.default
    #                 diff_property.default_changed = True

    #             # Verificando se houve alteração de PK
    #             if property_old.pk != property_new.pk:
    #                 persist_alter = True
    #                 diff_property.old_pk = property_old.pk
    #                 diff_property.new_pk = property_new.pk
    #                 diff_property.pk_changed = True

    #             # Verificando se houve alteração de precisão
    #             if property_old.max_length != property_new.max_length:
    #                 persist_alter = True
    #                 diff_property.new_max_length = property_new.max_length
    #                 diff_property.max_length_changed = True

    #             # Constroi representação do check do enumerado old
    #             old_hash = None
    #             if property_old.domain_config:
    #                 mapped_values_old = [
    #                     v.mapped_value for v in property_old.domain_config
    #                 ]
    #                 mapped_values_old.sort(key=lambda x: x)
    #                 # Fazendo um hash da lista
    #                 old_hash = hash(tuple(mapped_values_old))

    #             # Constroi representação do check do enumerado new
    #             new_hash = None
    #             mapped_values_new = []
    #             if property_new.domain_config:
    #                 mapped_values_new = [
    #                     v.mapped_value for v in property_new.domain_config
    #                 ]
    #                 mapped_values_new.sort(key=lambda x: x)
    #                 # Fazendo um hash da lista
    #                 new_hash = hash(tuple(mapped_values_new))

    #             # Verificando se houve alteração de enumerado
    #             if old_hash != new_hash:
    #                 persist_alter = True
    #                 diff_property.new_check = mapped_values_new
    #                 diff_property.check_changed = True

    #             # Guardando a alteração, caso necessário
    #             if persist_alter:
    #                 alter_properties.append(diff_property)

    # FIXME Comentado por hora (para não suportar ainda o DROP de colunas)
    # Iterando as propriedades do objeto OLD, e verificando as que não existem no objeto NEW
    # (considera-se aqui também a equilência por nome da propriedade ou da coluna)
    # for property_name_old in properties_structure_old.properties:
    #     if property_name_old not in mapa_equivalencias_old_new:
    #         entity_mapping_old = properties_structure_old.entity_properties.get(
    #             property_name_old
    #         )
    #         entity_name_old = property_name_old
    #         if entity_mapping_old and entity_mapping_old.column:
    #             entity_name_old = entity_mapping_old.column

    #         property_old = properties_structure_old.properties[property_name_old]
    #         drop_properties.append(
    #             DiffProperty(
    #                 action=DiffAction.DROP_COLUMN,
    #                 property_name=property_name_old,
    #                 entity_name=entity_name_old,
    #                 old_datatype=property_old.type,
    #                 old_pk=property_old.pk,
    #             )
    #         )

    # return alter_properties, add_properties, drop_properties

    def _build_column_and_fk_specs(
        self,
        entity_model: EntityModelBase,
        properties_structure: PropertiesCompilerStructure,
        entity_models: dict[str, EntityModel],
    ) -> tuple[list[dict[str, Any]], list[dict[str, str]]]:
        column_specs: list[dict[str, Any]] = []
        fk_specs: list[dict[str, str]] = []
        existing_columns: set[str] = set()

        for property_name, property_model in properties_structure.properties.items():
            if not isinstance(property_model.type, PrimitiveTypes):
                relation_ref = RelationRefParser.parse(property_model.type)
                target_model = relation_ref.resolve_model(entity_models) if relation_ref else None
                # Tenta identificar FK local a partir do relation_column
                entity_mapping = properties_structure.entity_properties.get(
                    property_name
                )
                relation_column = (
                    entity_mapping.relation_column if entity_mapping else None
                )
                if relation_column:
                    relation_column_ref, column_name = self._extract_relation_column_target(
                        relation_column, entity_model
                    )
                    # len >= 3 => <escopo>/<entidade>(/ ...)/<coluna>
                    if relation_column_ref and column_name:
                        # Coluna está na própria entidade
                        if relation_column_ref.target_id == entity_model.id:
                            column_spec = self._build_column_spec_for_relation(
                                property_name,
                                property_model,
                                column_name,
                                properties_structure,
                                target_model,
                            )
                            if column_spec["column_name"] not in existing_columns:
                                column_specs.append(column_spec)
                                existing_columns.add(column_spec["column_name"])
                            fk_specs.append(
                                self._build_fk_spec(
                                    entity_model,
                                    column_name,
                                    property_model,
                                    entity_models,
                                    target_model=target_model,
                                )
                            )
                continue
            entity_mapping = properties_structure.entity_properties.get(property_name)
            column_name = (
                entity_mapping.column
                if entity_mapping and entity_mapping.column
                else property_name
            )

            enum_values: list[Any] = []
            if property_model.domain_config:
                enum_values = [v.mapped_value for v in property_model.domain_config]
                enum_values.sort(key=lambda x: x)

            maximum_value = getattr(property_model, "maximum", None)

            column_specs.append(
                {
                    "property_name": property_name,
                    "column_name": column_name,
                    "datatype": property_model.type,
                    "sql_type": MigrationCompilerUtil.resolve_sql_type(
                        property_model.type, property_model.max_length
                    ),
                    "not_null": property_name in properties_structure.required,
                    "default": property_model.default,
                    "description": property_model.description,
                    "is_pk": bool(property_model.pk),
                    "enum_values": enum_values,
                    "max_length": property_model.max_length,
                    "maximum": maximum_value,
                    "is_numeric": MigrationCompilerUtil.is_numeric(property_model.type),
                }
            )

        # Verifica relacionamentos definidos em outras entidades que apontam colunas para a tabela atual
        for other_id, other_entity in entity_models.items():
            if other_id == f"{entity_model.escopo}/{entity_model.id}" or other_id == entity_model.id:
                continue
            for other_prop_name, other_prop_model in other_entity.properties.items():
                if isinstance(other_prop_model.type, PrimitiveTypes):
                    continue
                other_entity_mapping = other_entity.repository.properties.get(
                    other_prop_name
                )
                relation_column = (
                    other_entity_mapping.relation_column if other_entity_mapping else None
                )
                if not relation_column:
                    continue
                relation_column_ref, column_name = self._extract_relation_column_target(
                    relation_column, other_entity
                )
                if not relation_column_ref or not column_name:
                    continue
                if relation_column_ref.target_id != entity_model.id:
                    continue

                target_model = relation_column_ref.resolve_model(
                    entity_models, base_model=other_entity
                )
                reference_model = (
                    target_model
                    if target_model and target_model != entity_model
                    else other_entity
                )

                # A coluna mora na tabela atual, e referencia a entidade "dona" da propriedade (other_entity)
                pk_type = self._resolve_pk_type(reference_model)
                pk_column = self._resolve_pk_column(reference_model)

                if column_name not in existing_columns:
                    column_specs.append(
                        self._build_external_column_spec(column_name, pk_type)
                    )
                    existing_columns.add(column_name)
                fk_specs.append(
                    {
                        "column_name": column_name,
                        "ref_table": reference_model.repository.map,
                        "ref_column": pk_column,
                    }
                )

        return column_specs, fk_specs

    @staticmethod
    def _extract_relation_column_target(
        relation_column: str,
        base_model: EntityModelBase | None = None,
    ) -> tuple[RelationRef | None, str | None]:
        parts = [part for part in relation_column.split("/") if part]
        if len(parts) < 2:
            return None, None

        column_name = parts[-1]
        relation_path = "/".join(parts[:-1])

        relation_ref = RelationRefParser.parse(relation_path)
        if relation_ref:
            if not relation_ref.scope and base_model and getattr(base_model, "escopo", None):
                relation_ref.scope = base_model.escopo  # type: ignore
            return relation_ref, column_name

        # Fallback para o formato antigo <escopo>/<entidade>/<coluna>
        if len(parts) >= 3:
            return RelationRefParser.parse("/".join(parts[:-1])), column_name

        return None, column_name

    def _build_column_spec_for_relation(
        self,
        property_name: str,
        property_model: Any,
        local_column: str,
        properties_structure: PropertiesCompilerStructure,
        target_model: EntityModelBase | None,
    ) -> dict[str, Any]:
        # Descobre o tipo da PK da entidade referenciada
        pk_type = PrimitiveTypes.UUID
        if target_model:
            pk_prop = next(
                (p for p in target_model.properties if target_model.properties[p].pk),
                None,
            )
            if pk_prop:
                pk_type = target_model.properties[pk_prop].type

        enum_values: list[Any] = []
        maximum_value = None

        return {
            "property_name": property_name,
            "column_name": local_column,
            "datatype": pk_type,
            "sql_type": MigrationCompilerUtil.resolve_sql_type(pk_type, None),
            "not_null": property_name in properties_structure.required,
            "default": None,
            "description": None,
            "is_pk": False,
            "enum_values": enum_values,
            "max_length": None,
            "maximum": maximum_value,
            "is_numeric": MigrationCompilerUtil.is_numeric(pk_type),
        }

    def _build_fk_spec(
        self,
        entity_model: EntityModelBase,
        local_column: str,
        property_model: Any,
        entity_models: dict[str, EntityModel],
        target_model: EntityModelBase | None = None,
    ) -> dict[str, str]:
        ref_entity = target_model
        if ref_entity is None:
            ref_entity_id = property_model.type
            ref_entity = entity_models.get(ref_entity_id)
        ref_table = None
        ref_pk_column = None
        if ref_entity:
            ref_table = ref_entity.repository.map
            ref_pk_prop = next(
                (p for p in ref_entity.properties if ref_entity.properties[p].pk),
                None,
            )
            if ref_pk_prop:
                ref_pk_column = ref_pk_prop
                entity_mapping_ref = ref_entity.repository.properties.get(ref_pk_prop)
                if entity_mapping_ref and entity_mapping_ref.column:
                    ref_pk_column = entity_mapping_ref.column

        return {
            "column_name": local_column,
            "ref_table": ref_table or "",
            "ref_column": ref_pk_column or "id",
        }

    def _resolve_pk_type(self, entity_model: EntityModelBase) -> PrimitiveTypes:
        pk_prop = next(
            (p for p in entity_model.properties if entity_model.properties[p].pk), None
        )
        if not pk_prop:
            return PrimitiveTypes.UUID
        return entity_model.properties[pk_prop].type

    def _resolve_pk_column(self, entity_model: EntityModelBase) -> str:
        pk_prop = next(
            (p for p in entity_model.properties if entity_model.properties[p].pk), None
        )
        if not pk_prop:
            return "id"
        entity_mapping = entity_model.repository.properties.get(pk_prop)
        if entity_mapping and entity_mapping.column:
            return entity_mapping.column
        return pk_prop

    def _build_external_column_spec(
        self, column_name: str, pk_type: PrimitiveTypes
    ) -> dict[str, Any]:
        """
        Constrói especificação de coluna para FK definida em outra entidade,
        mas cuja coluna está na tabela atual.
        """
        return {
            "property_name": column_name,
            "column_name": column_name,
            "datatype": pk_type,
            "sql_type": MigrationCompilerUtil.resolve_sql_type(pk_type, None),
            "not_null": False,
            "default": None,
            "description": None,
            "is_pk": False,
            "enum_values": [],
            "max_length": None,
            "maximum": None,
            "is_numeric": MigrationCompilerUtil.is_numeric(pk_type),
        }

    def _detect_renamed_columns(
        self,
        properties_structure: PropertiesCompilerStructure,
        properties_structure_old: PropertiesCompilerStructure | None,
    ) -> list[tuple[str, str]]:
        """
        Retorna lista de tuplas (old_column, new_column) quando o mapeamento
        de uma propriedade mudou de nome de coluna.
        """
        if not properties_structure_old:
            return []

        rename_ops: list[tuple[str, str]] = []
        for prop_name, _ in properties_structure.properties.items():
            if prop_name not in properties_structure_old.properties:
                continue

            entity_mapping_new = properties_structure.entity_properties.get(prop_name)
            new_column = (
                entity_mapping_new.column
                if entity_mapping_new and entity_mapping_new.column
                else prop_name
            )

            entity_mapping_old = properties_structure_old.entity_properties.get(
                prop_name
            )
            old_column = (
                entity_mapping_old.column
                if entity_mapping_old and entity_mapping_old.column
                else prop_name
            )

            if old_column != new_column:
                rename_ops.append((old_column, new_column))

        return rename_ops



import json
import os
import sys
from pathlib import Path

import pytest
from pydantic import ValidationError

ROOT_DIR = Path(__file__).resolve().parents[1]
if str(ROOT_DIR) not in sys.path:
    sys.path.insert(0, str(ROOT_DIR))

os.environ.setdefault("ESCOPO_RESTLIB2", "test-scope")

from nsj_rest_lib2.compiler.compiler import EDLCompiler
from nsj_rest_lib2.compiler.edl_model.entity_model import EntityModel
from nsj_rest_lib2.compiler.edl_model.entity_model_root import EntityModelRoot
from nsj_rest_lib2.compiler.util.type_naming_util import (
    compile_function_class_name,
)
from nsj_rest_lib2.dto.escopo_dto import EscopoDTO


def _load_entity(path: Path):
    with path.open("r") as fp:
        edl_data = json.load(fp)

    model = (
        EntityModelRoot(**edl_data)
        if edl_data.get("mixin", False)
        else EntityModel(**edl_data)
    )
    complete_id = f"{model.escopo}/{model.id}"
    return complete_id, model


def _load_entity_models(base_dir: Path) -> dict[str, EntityModel]:
    models: dict[str, EntityModel] = {}
    for path in base_dir.iterdir():
        if path.suffix.lower() != ".json":
            continue
        key, model = _load_entity(path)
        models[key] = model
    return models


def test_compile_handlers_for_classificacao_financeira_generates_delete_function_type():
    base_dir = ROOT_DIR / "@schemas_test"
    entity_models = _load_entity_models(base_dir)

    model_key = "financas/ClassificacaoFinanceira"
    entity_model = entity_models[model_key]

    compiler = EDLCompiler()
    escopo = EscopoDTO(codigo=entity_model.escopo, service_account=None)
    result = compiler.compile_model(
        entity_model,
        list(entity_models.items()),
        escopo=escopo,
    )

    assert result is not None

    # Nomes das funções de banco para post/put/delete
    assert result.insert_function_name == "financas.api_classificacaofinanceiranovo"
    assert result.update_function_name == "financas.api_classificacaofinanceiraalterar"
    assert result.delete_function_name == "financas.api_classificacaofinanceiraexcluir"
    assert result.retrieve_after_insert is True
    assert result.retrieve_after_update is True
    assert result.retrieve_after_partial_update is False
    assert result.post_response_dto_class_name is None
    assert result.custom_json_post_response is False
    assert result.custom_json_put_response is False
    assert result.custom_json_patch_response is False

    # FunctionType para delete
    delete_code = result.source_delete_function_type
    assert delete_code is not None

    expected_class = compile_function_class_name(entity_model.id, "", [], "delete")
    assert result.delete_function_type_class_name == expected_class
    assert f"class {expected_class}" in delete_code
    # Campos esperados (derivados do mapping do handler de delete)
    assert "classificacao: uuid.UUID = FunctionField(" in delete_code
    assert "grupoempresarial: uuid.UUID = FunctionField(" in delete_code


def test_compile_handlers_generates_get_and_list_function_types_from_edl():
    edl_json = {
        "edl_version": "1.0",
        "escopo": "test",
        "description": "Entidade de teste para funções GET/LIST.",
        "id": "Foo",
        "version": "1.0",
        "properties": {
            "id": {"type": "uuid", "pk": True},
            "codigo": {"type": "string"},
        },
        "repository": {
            "map": "test.foo",
            "shared_table": False,
            "properties": {
                "id": {"column": "id"},
                "codigo": {"column": "codigo"},
            },
            "indexes": [],
        },
        "api": {
            "resource": "foos",
            "expose": True,
            "verbs": ["GET"],
            "handlers": {
                "get": {
                    "impl": "pg_function",
                    "function_ref": "test.fn_foo_get",
                    "call": {
                        "arg_binding": {
                            "type_name": "test.tfoo_get",
                            "mapping": [
                                {"attr": "id_func", "from": "path.id"},
                                {"attr": "tenant", "from": "args.tenant"},
                            ],
                        }
                    },
                    "result": {"expected": "entity_row"},
                },
                "list": {
                    "impl": "pg_function",
                    "function_ref": "test.fn_foo_list",
                    "call": {
                        "arg_binding": {
                            "type_name": "test.tfoo_list",
                            "mapping": [
                                {"attr": "tenant", "from": "args.tenant"},
                                {"attr": "search", "from": "args.search"},
                            ],
                        }
                    },
                    "result": {"expected": "entity_row"},
                },
            },
        },
    }

    compiler = EDLCompiler()
    escopo = EscopoDTO(codigo="test", service_account=None)
    result = compiler.compile_model_from_edl(edl_json, [], escopo=escopo)

    assert result is not None
    assert result.get_function_name == "test.fn_foo_get"
    assert result.list_function_name == "test.fn_foo_list"

    get_code = result.source_get_function_type
    list_code = result.source_list_function_type
    assert get_code is not None
    assert list_code is not None

    expected_get_class = compile_function_class_name("Foo", "", [], "get")
    expected_list_class = compile_function_class_name("Foo", "", [], "list")

    assert result.get_function_type_class_name == expected_get_class
    assert result.list_function_type_class_name == expected_list_class
    assert f"class {expected_get_class}" in get_code
    assert "id_func: uuid.UUID = FunctionField(" in get_code
    assert "tenant: Any = FunctionField(" in get_code

    assert f"class {expected_list_class}" in list_code
    assert "tenant: Any = FunctionField(" in list_code
    assert "search: Any = FunctionField(" in list_code
    assert result.custom_json_get_response is False
    assert result.custom_json_list_response is False


def test_compile_handlers_generates_partial_row_response_dto():
    edl_json = {
        "edl_version": "1.0",
        "escopo": "test",
        "description": "Entidade de teste para partial_row.",
        "id": "Bar",
        "version": "1.0",
        "properties": {
            "id": {"type": "uuid", "pk": True},
            "codigo": {"type": "string"},
            "descricao": {"type": "string"},
        },
        "repository": {
            "map": "test.bar",
            "shared_table": False,
            "properties": {
                "id": {"column": "id"},
                "codigo": {"column": "codigo"},
                "descricao": {"column": "descricao"},
            },
            "indexes": [],
        },
        "api": {
            "resource": "bars",
            "expose": True,
            "verbs": ["POST"],
            "handlers": {
                "post": {
                    "impl": "pg_function",
                    "function_ref": "test.fn_bar",
                    "call": {
                        "arg_binding": {
                            "type_name": "test.tbar",
                            "mapping": [{"attr": "codigo", "from": "body.codigo"}],
                        }
                    },
                    "result": {"expected": "partial_row", "properties": ["codigo"]},
                }
            },
        },
    }

    compiler = EDLCompiler()
    escopo = EscopoDTO(codigo="test", service_account=None)
    result = compiler.compile_model_from_edl(edl_json, [], escopo=escopo)

    assert result is not None
    assert result.post_response_dto_class_name == "BarPostResponseDTO"
    assert "class BarPostResponseDTO" in result.dto_code
    response_section = result.dto_code.split("class BarPostResponseDTO")[1]
    assert "codigo: str = DTOField" in response_section
    assert "id: uuid.UUID = DTOField" not in response_section
    assert result.retrieve_after_insert is False
    assert result.custom_json_post_response is False


def test_compile_handlers_sets_custom_json_flag_without_impl():
    edl_json = {
        "edl_version": "1.0",
        "escopo": "test",
        "description": "Entidade de teste para custom_json.",
        "id": "Baz",
        "version": "1.0",
        "properties": {
            "id": {"type": "uuid", "pk": True},
        },
        "repository": {
            "map": "test.baz",
            "shared_table": False,
            "properties": {
                "id": {"column": "id"},
            },
            "indexes": [],
        },
        "api": {
            "resource": "baz",
            "expose": True,
            "verbs": ["POST"],
            "handlers": {
                "post": {
                    "result": {"expected": "custom_json"},
                }
            },
        },
    }

    compiler = EDLCompiler()
    escopo = EscopoDTO(codigo="test", service_account=None)
    result = compiler.compile_model_from_edl(edl_json, [], escopo=escopo)

    assert result is not None
    assert result.custom_json_post_response is True
    assert result.retrieve_after_insert is False
    assert result.post_response_dto_class_name is None


def test_compile_handlers_sets_custom_json_for_get_and_list():
    edl_json = {
        "edl_version": "1.0",
        "escopo": "test",
        "description": "Entidade de teste para custom_json em GET/LIST.",
        "id": "FooBar",
        "version": "1.0",
        "properties": {
            "id": {"type": "uuid", "pk": True},
        },
        "repository": {
            "map": "test.foobar",
            "shared_table": False,
            "properties": {
                "id": {"column": "id"},
            },
            "indexes": [],
        },
        "api": {
            "resource": "foobars",
            "expose": True,
            "verbs": ["GET"],
            "handlers": {
                "get": {
                    "result": {"expected": "custom_json"},
                },
                "list": {
                    "result": {"expected": "custom_json"},
                },
            },
        },
    }

    compiler = EDLCompiler()
    escopo = EscopoDTO(codigo="test", service_account=None)
    result = compiler.compile_model_from_edl(edl_json, [], escopo=escopo)

    assert result is not None
    assert result.custom_json_get_response is True
    assert result.custom_json_list_response is True


def test_compile_handlers_sets_custom_json_for_delete():
    edl_json = {
        "edl_version": "1.0",
        "escopo": "test",
        "description": "Entidade de teste para custom_json em DELETE.",
        "id": "FooDelete",
        "version": "1.0",
        "properties": {
            "id": {"type": "uuid", "pk": True},
        },
        "repository": {
            "map": "test.foodelete",
            "shared_table": False,
            "properties": {
                "id": {"column": "id"},
            },
            "indexes": [],
        },
        "api": {
            "resource": "foodeletes",
            "expose": True,
            "verbs": ["DELETE"],
            "handlers": {
                "delete": {
                    "result": {"expected": "custom_json"},
                }
            },
        },
    }

    compiler = EDLCompiler()
    escopo = EscopoDTO(codigo="test", service_account=None)
    result = compiler.compile_model_from_edl(edl_json, [], escopo=escopo)

    assert result is not None
    assert result.custom_json_delete_response is True


def test_compile_handlers_retrieve_after_update_and_partial_update_flags():
    edl_json = {
        "edl_version": "1.0",
        "escopo": "test",
        "description": "Entidade de teste para flags de retrieve em PUT/PATCH.",
        "id": "RetrieveFlags",
        "version": "1.0",
        "properties": {
            "id": {"type": "uuid", "pk": True},
            "codigo": {"type": "string"},
        },
        "repository": {
            "map": "test.retrieve_flags",
            "shared_table": False,
            "properties": {
                "id": {"column": "id"},
                "codigo": {"column": "codigo"},
            },
            "indexes": [],
        },
        "api": {
            "resource": "retrieve-flags",
            "expose": True,
            "verbs": ["PUT", "PATCH"],
            "handlers": {
                "put": {
                    "result": {"expected": "entity_row"},
                },
                "patch": {
                    "result": {"expected": "entity_row"},
                },
            },
        },
    }

    compiler = EDLCompiler()
    escopo = EscopoDTO(codigo="test", service_account=None)
    result = compiler.compile_model_from_edl(edl_json, [], escopo=escopo)

    assert result is not None
    assert result.retrieve_after_update is True
    assert result.retrieve_after_partial_update is True


def test_compile_handlers_disable_retrieve_after_update_for_partial_row():
    edl_json = {
        "edl_version": "1.0",
        "escopo": "test",
        "description": "Entidade de teste para partial_row em PUT/PATCH.",
        "id": "PartialRetrieveFlags",
        "version": "1.0",
        "properties": {
            "id": {"type": "uuid", "pk": True},
            "codigo": {"type": "string"},
        },
        "repository": {
            "map": "test.partial_retrieve_flags",
            "shared_table": False,
            "properties": {
                "id": {"column": "id"},
                "codigo": {"column": "codigo"},
            },
            "indexes": [],
        },
        "api": {
            "resource": "partial-retrieve-flags",
            "expose": True,
            "verbs": ["PUT", "PATCH"],
            "handlers": {
                "put": {
                    "result": {"expected": "partial_row", "properties": ["codigo"]},
                },
                "patch": {
                    "result": {"expected": "partial_row", "properties": ["codigo"]},
                },
            },
        },
    }

    compiler = EDLCompiler()
    escopo = EscopoDTO(codigo="test", service_account=None)
    result = compiler.compile_model_from_edl(edl_json, [], escopo=escopo)

    assert result is not None
    assert result.retrieve_after_update is False
    assert result.retrieve_after_partial_update is False


def test_compile_handlers_partial_row_requires_properties():
    edl_json = {
        "edl_version": "1.0",
        "escopo": "test",
        "description": "Entidade inválida para partial_row.",
        "id": "Qux",
        "version": "1.0",
        "properties": {
            "id": {"type": "uuid", "pk": True},
        },
        "repository": {
            "map": "test.qux",
            "shared_table": False,
            "properties": {
                "id": {"column": "id"},
            },
            "indexes": [],
        },
        "api": {
            "resource": "qux",
            "expose": True,
            "verbs": ["POST"],
            "handlers": {
                "post": {
                    "result": {"expected": "partial_row"},
                }
            },
        },
    }

    compiler = EDLCompiler()
    escopo = EscopoDTO(codigo="test", service_account=None)

    with pytest.raises(ValidationError):
        compiler.compile_model_from_edl(edl_json, [], escopo=escopo)



import uuid

from nsj_rest_lib.entity.entity_base import EntityBase
from nsj_rest_lib.decorator.entity import Entity


@Entity(
    table_name="restlib2.escopo",
    pk_field="id",
    default_order_fields=["codigo", "id"],
)
class EscopoEntity(EntityBase):
    id: uuid.UUID = None
    codigo: str = None
    service_account: str = None



import json

from typing import Any, Callable

from flask import Flask, request

from nsj_rest_lib.settings import APP_NAME

from nsj_gcf_utils.rest_error_util import format_json_error

from nsj_multi_database_lib.decorator.multi_database import multi_database

from nsj_rest_lib.controller.controller_util import DEFAULT_RESP_HEADERS
from nsj_rest_lib.controller.delete_route import DeleteRoute
from nsj_rest_lib.controller.get_route import GetRoute
from nsj_rest_lib.controller.list_route import ListRoute
from nsj_rest_lib.controller.patch_route import PatchRoute
from nsj_rest_lib.controller.post_route import PostRoute
from nsj_rest_lib.controller.put_route import PutRoute

from nsj_rest_lib2.exception import MissingEntityConfigException
from nsj_rest_lib2.service.entity_loader import EntityLoader


def _get_query_args() -> tuple[str, str, bool]:
    # Tentando ler do query args
    query_args = request.args
    tenant = query_args.get("tenant")
    grupo_empresarial = query_args.get("grupo_empresarial")
    force_reload = query_args.get("force_reload", "false").lower() == "true"

    # Tentando ler do corpo da requisição
    try:
        body_str = request.data.decode("utf-8")
        body_json = json.loads(body_str)

        if not tenant:
            tenant = body_json.get("tenant")
        if not grupo_empresarial:
            grupo_empresarial = body_json.get("grupo_empresarial")
    except:
        pass

    return (str(tenant), str(grupo_empresarial), force_reload)


def _endpoint_name(func: Any, multidb: bool, root: str) -> str:
    suffix = "_mb" if multidb else ""
    return f"{root}_{func.__name__}{suffix}"


def setup_dynamic_routes(
    flask_app: Flask,
    multidb: bool = True,
    dynamic_root_path: str = "edl1",
    injector_factory: Any = None,
    escopo_in_url: bool = False,
    edls_path: str | None = None,
) -> None:

    if not escopo_in_url:
        COLLECTION_DYNAMIC_ROUTE = f"/{APP_NAME}/{dynamic_root_path}/<entity_resource>"
        ONE_DYNAMIC_ROUTE = f"/{APP_NAME}/{dynamic_root_path}/<entity_resource>/<id>"
    else:
        COLLECTION_DYNAMIC_ROUTE = (
            f"/{APP_NAME}/{dynamic_root_path}/<entity_escopo>/<entity_resource>"
        )
        ONE_DYNAMIC_ROUTE = (
            f"/{APP_NAME}/{dynamic_root_path}/<entity_escopo>/<entity_resource>/<id>"
        )

    entity_loader = EntityLoader(edls_path=edls_path)

    def _dynamic_route_wrapper(
        route_builder: Callable[[tuple[Any, ...]], Callable[..., Any]],
        endpoint_suffix: str | None = None,
    ) -> Callable[..., Any]:
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            if "entity_resource" not in kwargs:
                msg = "Faltando parâmetro identificador da entidade na URL."
                return (format_json_error(msg), 400, {**DEFAULT_RESP_HEADERS})
            entity_resource = kwargs.pop("entity_resource")

            entity_escopo = kwargs.pop("entity_escopo", "")

            tenant, grupo_empresarial, force_reload = _get_query_args()

            try:
                entity_config = entity_loader.load_entity_source(
                    entity_resource,
                    tenant,
                    grupo_empresarial,
                    escopo=entity_escopo,
                    force_reload=force_reload,
                )
            except MissingEntityConfigException:
                msg = f"Entity configuration for {entity_resource} not found."
                return (format_json_error(msg), 412, {**DEFAULT_RESP_HEADERS})

            (
                _,
                _,
                _,
                _,
                _,
                service_account,
                *rest,
            ) = entity_config

            route_handler = route_builder(entity_config)

            if multidb:
                route_handler = multi_database(service_account)(route_handler)

            return route_handler(*args, **kwargs)

        wrapper.__name__ = endpoint_suffix or f"{route_builder.__name__}_wrapper"
        return wrapper

    def list_dynamic_builder(entity_config: tuple[Any, ...]) -> Callable[..., Any]:
        (
            dto_class_name,
            entity_class_name,
            etities_dict,
            api_expose,
            api_verbs,
            _service_account,
            _insert_function_class_name,
            _update_function_class_name,
            _insert_function_name,
            _update_function_name,
            _get_function_name,
            list_function_name,
            _delete_function_name,
            _get_function_type_class_name,
            list_function_type_class_name,
            _delete_function_type_class_name,
            *_,
            _custom_json_get_response,
            custom_json_list_response,
            _custom_json_delete_response,
        ) = entity_config

        def list_dynamic(*args: Any, **kwargs: Any) -> Any:
            if not api_expose or "GET" not in api_verbs:
                return ("", 405, {})

            route = ListRoute(
                url=COLLECTION_DYNAMIC_ROUTE,
                http_method="GET",
                dto_class=etities_dict[dto_class_name],
                entity_class=etities_dict[entity_class_name],
                injector_factory=injector_factory,
                list_function_name=list_function_name,
                list_function_type_class=etities_dict.get(
                    list_function_type_class_name
                ),
                custom_json_response=bool(custom_json_list_response),
            )

            return route.internal_handle_request(*args, **kwargs)

        return list_dynamic

    def get_dynamic_builder(entity_config: tuple[Any, ...]) -> Callable[..., Any]:
        (
            dto_class_name,
            entity_class_name,
            etities_dict,
            api_expose,
            api_verbs,
            _service_account,
            _insert_function_class_name,
            _update_function_class_name,
            _insert_function_name,
            _update_function_name,
            get_function_name,
            _list_function_name,
            _delete_function_name,
            get_function_type_class_name,
            _list_function_type_class_name,
            _delete_function_type_class_name,
            *_,
            custom_json_get_response,
            _custom_json_list_response,
            _custom_json_delete_response,
        ) = entity_config

        def get_dynamic(*args: Any, **kwargs: Any) -> Any:
            if not api_expose or "GET" not in api_verbs:
                return ("", 405, {})

            route = GetRoute(
                url=ONE_DYNAMIC_ROUTE,
                http_method="GET",
                dto_class=etities_dict[dto_class_name],
                entity_class=etities_dict[entity_class_name],
                injector_factory=injector_factory,
                get_function_name=get_function_name,
                get_function_type_class=etities_dict.get(get_function_type_class_name),
                custom_json_response=bool(custom_json_get_response),
            )

            return route.internal_handle_request(*args, **kwargs)

        return get_dynamic

    def post_dynamic_builder(entity_config: tuple[Any, ...]) -> Callable[..., Any]:
        (
            dto_class_name,
            entity_class_name,
            etities_dict,
            api_expose,
            api_verbs,
            _service_account,
            insert_function_class_name,
            _update_function_class_name,
            insert_function_name,
            _update_function_name,
            _get_function_name,
            _list_function_name,
            _delete_function_name,
            _get_function_type_class_name,
            _list_function_type_class_name,
            _delete_function_type_class_name,
            retrieve_after_insert,
            _retrieve_after_update,
            _retrieve_after_partial_update,
            post_response_dto_class_name,
            _put_response_dto_class_name,
            _patch_response_dto_class_name,
            custom_json_post_response,
            _custom_json_put_response,
            _custom_json_patch_response,
            _custom_json_get_response,
            _custom_json_list_response,
            _custom_json_delete_response,
        ) = entity_config

        def post_dynamic(*args: Any, **kwargs: Any) -> Any:
            if not api_expose or "POST" not in api_verbs:
                return ("", 405, {})

            insert_function_type_class = (
                etities_dict.get(insert_function_class_name)
                if insert_function_class_name
                else None
            )

            dto_response_class = (
                etities_dict.get(post_response_dto_class_name)
                if post_response_dto_class_name
                else None
            )

            route = PostRoute(
                url=COLLECTION_DYNAMIC_ROUTE,
                http_method="POST",
                dto_class=etities_dict[dto_class_name],
                entity_class=etities_dict[entity_class_name],
                injector_factory=injector_factory,
                insert_function_type_class=insert_function_type_class,
                insert_function_name=insert_function_name,
                dto_response_class=dto_response_class,
                retrieve_after_insert=retrieve_after_insert,
                custom_json_response=bool(custom_json_post_response),
            )

            return route.internal_handle_request(*args, **kwargs)

        return post_dynamic

    def put_dynamic_builder(entity_config: tuple[Any, ...]) -> Callable[..., Any]:
        (
            dto_class_name,
            entity_class_name,
            etities_dict,
            api_expose,
            api_verbs,
            _service_account,
            _insert_function_class_name,
            update_function_class_name,
            _insert_function_name,
            update_function_name,
            _get_function_name,
            _list_function_name,
            _delete_function_name,
            _get_function_type_class_name,
            _list_function_type_class_name,
            _delete_function_type_class_name,
            _retrieve_after_insert,
            retrieve_after_update,
            _retrieve_after_partial_update,
            _post_response_dto_class_name,
            put_response_dto_class_name,
            _patch_response_dto_class_name,
            _custom_json_post_response,
            custom_json_put_response,
            _custom_json_patch_response,
            _custom_json_get_response,
            _custom_json_list_response,
            _custom_json_delete_response,
        ) = entity_config

        dto_response_class = (
            etities_dict.get(put_response_dto_class_name)
            if put_response_dto_class_name
            else None
        )

        def put_dynamic(*args: Any, **kwargs: Any) -> Any:
            if not api_expose or "PUT" not in api_verbs:
                return ("", 405, {})

            update_function_type_class = (
                etities_dict.get(update_function_class_name)
                if update_function_class_name
                else None
            )

            route = PutRoute(
                url=ONE_DYNAMIC_ROUTE,
                http_method="PUT",
                dto_class=etities_dict[dto_class_name],
                entity_class=etities_dict[entity_class_name],
                injector_factory=injector_factory,
                update_function_type_class=update_function_type_class,
                update_function_name=update_function_name,
                dto_response_class=dto_response_class,
                retrieve_after_update=retrieve_after_update,
                custom_json_response=bool(custom_json_put_response),
            )

            return route.internal_handle_request(*args, **kwargs)

        return put_dynamic

    def patch_dynamic_builder(entity_config: tuple[Any, ...]) -> Callable[..., Any]:
        (
            dto_class_name,
            entity_class_name,
            etities_dict,
            api_expose,
            api_verbs,
            _service_account,
            _insert_function_class_name,
            _update_function_class_name,
            _insert_function_name,
            _update_function_name,
            _get_function_name,
            _list_function_name,
            _delete_function_name,
            _get_function_type_class_name,
            _list_function_type_class_name,
            _delete_function_type_class_name,
            _retrieve_after_insert,
            _retrieve_after_update,
            retrieve_after_partial_update,
            _post_response_dto_class_name,
            _put_response_dto_class_name,
            patch_response_dto_class_name,
            _custom_json_post_response,
            _custom_json_put_response,
            custom_json_patch_response,
            _custom_json_get_response,
            _custom_json_list_response,
            _custom_json_delete_response,
        ) = entity_config

        def patch_dynamic(*args: Any, **kwargs: Any) -> Any:
            if not api_expose or "PATCH" not in api_verbs:
                return ("", 405, {})

            dto_response_class = (
                etities_dict.get(patch_response_dto_class_name)
                if patch_response_dto_class_name
                else None
            )

            route = PatchRoute(
                url=ONE_DYNAMIC_ROUTE,
                http_method="PATCH",
                dto_class=etities_dict[dto_class_name],
                entity_class=etities_dict[entity_class_name],
                injector_factory=injector_factory,
                dto_response_class=dto_response_class,
                retrieve_after_partial_update=retrieve_after_partial_update,
                custom_json_response=bool(custom_json_patch_response),
            )

            return route.internal_handle_request(*args, **kwargs)

        return patch_dynamic

    def delete_dynamic_builder(entity_config: tuple[Any, ...]) -> Callable[..., Any]:
        (
            dto_class_name,
            entity_class_name,
            etities_dict,
            api_expose,
            api_verbs,
            _service_account,
            _insert_function_class_name,
            _update_function_class_name,
            _insert_function_name,
            _update_function_name,
            _get_function_name,
            _list_function_name,
            delete_function_name,
            _get_function_type_class_name,
            _list_function_type_class_name,
            delete_function_type_class_name,
            *_,
            _custom_json_get_response,
            _custom_json_list_response,
            custom_json_delete_response,
        ) = entity_config

        def delete_dynamic(*args: Any, **kwargs: Any) -> Any:
            if not api_expose or "DELETE" not in api_verbs:
                return ("", 405, {})

            route = DeleteRoute(
                url=ONE_DYNAMIC_ROUTE,
                http_method="DELETE",
                dto_class=etities_dict[dto_class_name],
                entity_class=etities_dict[entity_class_name],
                injector_factory=injector_factory,
                delete_function_name=delete_function_name,
                delete_function_type_class=etities_dict.get(
                    delete_function_type_class_name
                ),
                custom_json_response=bool(custom_json_delete_response),
            )

            return route.internal_handle_request(*args, **kwargs)

        return delete_dynamic

    list_dynamic = _dynamic_route_wrapper(
        list_dynamic_builder, endpoint_suffix="list_dynamic"
    )
    get_dynamic = _dynamic_route_wrapper(
        get_dynamic_builder, endpoint_suffix="get_dynamic"
    )
    post_dynamic = _dynamic_route_wrapper(
        post_dynamic_builder, endpoint_suffix="post_dynamic"
    )
    put_dynamic = _dynamic_route_wrapper(
        put_dynamic_builder, endpoint_suffix="put_dynamic"
    )
    patch_dynamic = _dynamic_route_wrapper(
        patch_dynamic_builder, endpoint_suffix="patch_dynamic"
    )
    delete_dynamic = _dynamic_route_wrapper(
        delete_dynamic_builder, endpoint_suffix="delete_dynamic"
    )

    # Registrando as rotas no flask
    flask_app.add_url_rule(
        COLLECTION_DYNAMIC_ROUTE,
        endpoint=_endpoint_name(list_dynamic, multidb, dynamic_root_path),
        view_func=list_dynamic,
        methods=["GET"],
    )
    flask_app.add_url_rule(
        ONE_DYNAMIC_ROUTE,
        endpoint=_endpoint_name(get_dynamic, multidb, dynamic_root_path),
        view_func=get_dynamic,
        methods=["GET"],
    )
    flask_app.add_url_rule(
        COLLECTION_DYNAMIC_ROUTE,
        endpoint=_endpoint_name(post_dynamic, multidb, dynamic_root_path),
        view_func=post_dynamic,
        methods=["POST"],
    )
    flask_app.add_url_rule(
        ONE_DYNAMIC_ROUTE,
        endpoint=_endpoint_name(put_dynamic, multidb, dynamic_root_path),
        view_func=put_dynamic,
        methods=["PUT"],
    )
    flask_app.add_url_rule(
        ONE_DYNAMIC_ROUTE,
        endpoint=_endpoint_name(patch_dynamic, multidb, dynamic_root_path),
        view_func=patch_dynamic,
        methods=["PATCH"],
    )
    flask_app.add_url_rule(
        ONE_DYNAMIC_ROUTE,
        endpoint=_endpoint_name(delete_dynamic, multidb, dynamic_root_path),
        view_func=delete_dynamic,
        methods=["DELETE"],
    )
